from .arguments import (Empty, ArgumentKind, Argument, ArgumentAddress)
from .model import (
    Workflow, FunctionNode, NodeData, get_workflow, is_workflow,
    is_node_ready)
from .mutations import (reset_workflow, insert_result)
from .create import (from_call)
from .graphs import (invert_links)

__all__ = ['invert_links', 'from_call',
           'Workflow', 'FunctionNode', 'NodeData',
           'get_workflow', 'is_workflow', 'reset_workflow',
           'insert_result', 'Empty',
           'Argument', 'ArgumentAddress', 'ArgumentKind']



from pytest import raises
from noodles.workflow import (
    Empty, ArgumentAddress,
    ArgumentKind, is_workflow, get_workflow, Workflow)
from noodles import run_single, schedule, gather


def dummy(a, b, c, *args, **kwargs):
    pass


def test_is_workflow():
    assert is_workflow(Workflow(root=None, nodes={}, links={}))


def test_get_workflow():
    assert get_workflow(4) is None


@schedule
def value(a):
    return a


@schedule
def add(a, b):
    return a+b


@schedule
def sub(a, b):
    return a - b


def test_private():
    a = add(1, 1)
    a._private = 3
    assert a._private == 3
    assert not hasattr(run_single(a), '_private')


def test_merge_workflow():
    A = value(1)
    B = value(2)
    C = add(A, B)

    assert is_workflow(C)
    C = get_workflow(C)
    A = get_workflow(A)
    B = get_workflow(B)
    assert C.nodes[C.root].bound_args.args == (Empty, Empty)
    assert (C.root, ArgumentAddress(ArgumentKind.regular, 'a', None)) \
        in C.links[A.root]
    assert (C.root, ArgumentAddress(ArgumentKind.regular, 'b', None)) \
        in C.links[B.root]


def test_binder():
    A = value(1)
    B = value(2)
    C = gather(A, B)

    C = get_workflow(C)
    A = get_workflow(A)
    B = get_workflow(B)

    assert is_workflow(C)
    assert C.nodes[C.root].bound_args.args == (Empty, Empty)
    assert (C.root, ArgumentAddress(ArgumentKind.variadic, 'a', 0)) \
        in C.links[A.root]
    assert (C.root, ArgumentAddress(ArgumentKind.variadic, 'a', 1)) \
        in C.links[B.root]


@schedule
def takes_keywords(s, **kwargs):
    return s


def test_with_keywords():
    A = value(1)
    B = value(2)
    C = takes_keywords(a=A, b=B, s="regular!")
    C = get_workflow(C)
    A = get_workflow(A)
    B = get_workflow(B)

    assert is_workflow(C)
    assert C.nodes[C.root].bound_args.args == ("regular!",)
    assert C.nodes[C.root].bound_args.kwargs == {'a': Empty, 'b': Empty}


class Normal:
    pass


@schedule
class Scheduled:
    pass


def test_arg_by_ref():
    n = Normal()
    s = Scheduled()

    n.x = 4
    s.x = n
    n.x = 5
    s.y = n

    result = run_single(s)
    assert result.x.x == 4
    assert result.y.x == 5


def test_hidden_promise():
    with raises(TypeError):
        a = Normal()
        b = Scheduled()
        c = Scheduled()

        a.x = b
        c.x = a


def test_tuple_unpack():
    a = Scheduled()
    b = Scheduled()

    a.x, a.y = 2, 3
    b.x, b.y = sub(a.x, a.y), sub(a.y, a.x)

    result = run_single(b)
    assert result.x == -1
    assert result.y == 1



# from ..logger import log
from ...utility import object_name
from noodles import serial
# from .hybrid import hybrid_threaded_worker

# import random
import uuid
import xenon
import os
import sys


class XenonConfig:
    """Configuration to the Xenon library.

    Xenon is a Java library that offers a uniform interface to execute jobs.
    These jobs may be run locally, over ssh ar against a queue manager like
    SLURM.

    [Documentation to Xenon can be found online](http://nlesc.github.io/Xenon)

    :param name:
        The quasi human readable name to give to this Xenon instance.
        This defaults to a generated UUID.

    :param jobs_scheme:
        The scheme by which to schedule jobs. Should be one of 'local', 'ssh',
        'slurm' etc. See the Xenon documentation.

    :param files_scheme:
        The scheme by which to transfer files. Should be 'local' or 'ssh'.
        See the Xenon documentation.

    :param location:
        A location. This can be the host of the 'ssh' or 'slurm' server.

    :param credential:
        To enter a server through ssh, we need to have some credentials.
        Preferably, you have a private/public key pair by which you can
        identify yourself. Otherwise, this would be a combination of
        username/password. This functions that can create a credential
        object can be found in Xenon.credentials in the Xenon documentation.

    :param jobs_properties:
        Configuration to the Xenon.jobs module.

    :param files_properties:
        Configuration to the Xenon.files module.
    """
    def __init__(self, *, name=None, jobs_scheme='local', files_scheme='local',
                 location=None, credential=None, jobs_properties=None,
                 files_properties=None):
        self.name = name or ("xenon-" + str(uuid.uuid4()))
        self.jobs_scheme = jobs_scheme
        self.files_scheme = files_scheme
        self.location = location
        self.credential = credential
        self.jobs_properties = jobs_properties
        self.files_properties = files_properties

    @property
    def scheduler_args(self):
        properties = xenon.conversions.dict_to_HashMap(self.jobs_properties)
        return (self.jobs_scheme, self.location,
                self.credential, properties)

    @property
    def filesystem_args(self):
        return (self.files_scheme, self.location,
                self.credential, self.files_properties)


class RemoteJobConfig(object):
    """Configuration for a single remote Job.

    :param name:
        A quasi human recognizable name for this job. This will default to
        'remote-xxxxxx' where 'xxxxx' is some UUID.

    :param working_dir:
        The work directory where the job runs. Defaults to the current working
        directory.

    :param prefix:
        The path prefix of the Python system we're running on. This defaults to
        `sys.prefix`. If we are in a VirtualEnv, this means that the spawned
        job will run in the same VirtualEnv.

    :param exec_command:
        The command that is being executed. This defaults to `worker.sh` which
        should be located in the `working_dir`. This default script initialises
        the VirtualEnv en starts Python with `-m noodles.worker`, acting as a
        pilot job.

    :param init:
        You may specify a function that needs to be run before any jobs are
        being executed. There exist frameworks in the wild, which won't
        function otherwise.

    :param finish:
        This function may do some clean up, after all jobs have been done.

    :param verbose:
        Be verbose about what we're doing. This is for debugging purposes only.

    :param time_out:
        It may take a while before a job is actually running. This specifies
        the time to wait before giving up.
    """
    def __init__(self, *, name=None, working_dir=None, prefix=None,
                 exec_command=None, n_threads=1, registry=serial.base,
                 init=None, finish=None, verbose=False,
                 queue=None, time_out=5000):
        self.name = name or ("remote-" + str(uuid.uuid4()))
        self.working_dir = working_dir or os.getcwd()
        self.prefix = prefix or sys.prefix
        self.exec_command = exec_command
        self.n_threads = n_threads
        self.registry = registry
        self.init = init
        self.finish = finish
        self.verbose = verbose
        self.queue = queue
        self.time_out = time_out

    def command_line(self):
        cmd = ['/bin/bash',
               self.working_dir + '/worker.sh',
               self.prefix, 'online', '-name', self.name,
               '-registry', object_name(self.registry)]

        if self.init:
            cmd.extend(["-init", object_name(self.init)])
            # cmd.append("-init")

        if self.finish:
            cmd.extend(["-finish", object_name(self.finish)])
            # cmd.append("-finish")

        if self.verbose:
            cmd.append("-verbose")

        return cmd


class XenonJob:
    """Representative of a XenonJob.

    :param keeper:
        The XenonKeeper object.

    :param job:
        The internal job object.

    :param desc:
        The job description.
    """
    def __init__(self, keeper, job, desc):
        self.keeper = keeper
        self.job = job
        self.desc = desc

        if self.interactive:
            self.streams = self.get_streams()

    def wait_until_running(self, timeout):
        status = self.keeper.jobs.waitUntilRunning(
            self.job, timeout)
        return status

    def wait_until_done(self, timeout=0):
        status = self.keeper.jobs.waitUntilDone(
            self.job, timeout)
        return status

    def get_streams(self):
        return self.keeper.jobs.getStreams(self.job)

    @property
    def interactive(self):
        return self.job.isInteractive()


class XenonKeeper:
    is_initialized = False

    def __init__(self, log_level='ERROR'):
        if not XenonKeeper.is_initialized:
            xenon.init(log_level=log_level)  # noqa
            XenonKeeper.is_initialized = True

        self._x = xenon.Xenon()
        self.jobs = self._x.jobs()
        self.credentials = self._x.credentials()
        self._schedulers = {}

    def add_scheduler(self, config: XenonConfig):
        name = config.name
        self._schedulers[name] = XenonScheduler(self, config)
        return name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._x.close()


class XenonScheduler:
    def __init__(self, keeper, config):
        self.config = config
        self._x = keeper
        self._s = keeper.jobs.newScheduler(*config.scheduler_args)

    def submit(self, command, *, queue=None, interactive=True):
        desc = xenon.jobs.JobDescription()
        if interactive:
            desc.setInteractive(True)
        desc.setExecutable(command[0])
        desc.setArguments(command[1:])
        if queue:
            desc.setQueueName(queue)
        elif self.config.jobs_scheme == 'local':
            desc.setQueueName('multi')

        job = self._x.jobs.submitJob(self._s, desc)
        return XenonJob(self._x, job, desc)



# from .termapp import TermApp


class Display:
    def __init__(self):
        pass

    def report(self):
        pass

    def __call__(self, key, status, data, err_msg):
        getattr(self, status)(key, data, err_msg)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            if exc_type is KeyboardInterrupt:
                self.out << "\n" << ['fg', 255, 200, 50] \
                         << "User interrupt detected, abnormal exit.\n" \
                         << ['reset']
                return True

            if exc_type is SystemExit:
                return False

            print("Internal error encountered. Contact the developers: \n",
                  exc_type, exc_val)
            return False

        self.report()



from noodles import schedule, run_process, serial


# @schedule
def init():
    global s
    s = "This global variable needs to be here!"
    return True


# @schedule
def finish():
    return "Finish functino was run!"


@schedule
def checker():
    return s == "This global variable needs to be here!"


def test_globals():
    a = checker()
    result = run_process(a, n_processes=1, registry=serial.base,
                         init=init, finish=finish)
    assert result



from .registry import (Registry, Serialiser)
from ..utility import (object_name, look_up)


class SerNamedTuple(Serialiser):
    def __init__(self, cls):
        super(SerNamedTuple, self).__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(tuple(obj))

    def decode(self, cls, data):
        return cls(*data)


class SerAutoNamedTuple(Serialiser):
    def __init__(self):
        super(SerAutoNamedTuple, self).__init__('<namedtuple>')

    def encode(self, obj, make_rec):
        return make_rec({
            'name': object_name(type(obj)),
            'data': tuple(obj)})

    def decode(self, cls, data):
        return look_up(data['name'])(*data['data'])


def is_namedtuple(obj):
    return isinstance(obj, tuple) and hasattr(obj, '_fields')


def namedtuple_hook(obj):
    if is_namedtuple(obj):
        return '<namedtuple>'
    else:
        return None


def registry():
    return Registry(
        hooks={
            '<namedtuple>': SerAutoNamedTuple()
        },
        hook_fn=namedtuple_hook)



from ..workflow import (Workflow, is_node_ready, Empty)
from ..workflow.arguments import (serialize_arguments, ref_argument)
from ..serial import (Registry)
from .prov import (prov_key)


def links(wf, i, deps):
    for d in deps:
        for l in wf.links[d]:
            if l[0] == i:
                yield (l[1], wf.nodes[d].prov)


def empty_args(n):
    for arg in serialize_arguments(n.bound_args):
        if ref_argument(n.bound_args, arg) == Empty:
            yield arg


def set_global_provenance(wf: Workflow, registry: Registry):
    """Compute a global provenance key for the entire workflow
    before evaluation. This key can be used to store and retrieve
    results in a database. The key computed in this stage is different
    from the (local) provenance key that can be computed for a node
    if all its arguments are known.

    In cases where a result derives from other results that were
    computed in child workflows, we can prevent the workflow system
    from reevaluating the results at each step to find that we already
    had the end-result somewhere. This is where the global prov-key
    comes in.

    Each node is assigned a `prov` attribute. If all arguments for this
    node are known, this key will be the same as the local prov-key.
    If some of the arguments are still empty, we add the global prov-keys
    of the dependent nodes to the hash.

    In this algorithm we traverse from the bottom of the DAG to the top
    and back using a stack. This allows us to compute the keys for each
    node without modifying the node other than setting the `prov` attribute
    with the resulting key."""
    stack = [wf.root]

    while stack:
        i = stack.pop()
        n = wf.nodes[i]

        if n.prov:
            continue

        if is_node_ready(n):
            job_msg = registry.deep_encode(n)
            n.prov = prov_key(job_msg)
            continue

        deps = wf.inverse_links[i]
        todo = [j for j in deps if not wf.nodes[j].prov]

        if not todo:
            link_dict = dict(links(wf, i, deps))
            link_prov = registry.deep_encode(
                [link_dict[arg] for arg in empty_args(n)])
            job_msg = registry.deep_encode(n)
            n.prov = prov_key(job_msg, link_prov)
            continue

        stack.append(i)
        stack.extend(deps)



from noodles import (
    schedule, has_scheduled_methods, run_single,
    run_process, serial)


def registry():
    return serial.pickle() + serial.base()


@has_scheduled_methods
class A:
    def __init__(self, x):
        super(A, self).__init__()
        self.x = x

    @schedule
    def __call__(self, y):
        return self.x * y


def test_class_methods_00():
    a = A(7)
    b = a(6)

    result = run_single(b)
    assert result == 42


def test_class_methods_01():
    a = A(7)
    b = a(6)

    result = run_process(b, 1, registry)
    assert result == 42



"""
|▛▀▜|▙▄▟|▗▄▖|▝▀▘|
⌮⌬
○◉□▣◔◑◕●♾⚝☺☹✨✩❄❅❆✾⚛✗✓✔✘
"""

# import sys

from .termapp import TermApp


class Turtle:
    dx = [1, 0, -1, 0]
    dy = [0, -1, 0, 1]

    def __init__(self, x, y, d):
        self.jump(x, y, d)
        self.stack = []

    def turn(self, r):
        self.d = (self.d + r) % 4

    def move(self):
        self.x += Turtle.dx[self.d]
        self.y += Turtle.dy[self.d]

    def push(self):
        self.stack.append((self.x, self.y, self.d))

    def pop(self):
        self.x, self.y, self.d = self.stack[-1]
        del self.stack[-1]

    def jump(self, x, y, d):
        self.x = x
        self.y = y
        self.d = d


class Canvas:
    def __init__(self, width, height, turtle=None):
        self.width = width
        self.height = height
        self.turtle = Turtle(0, 0, 0) if turtle is None else Turtle(*turtle)
        self.field = [[0 for i in range(width)] for j in range(height)]

    def put(self, s):
        for c in s:
            if c == '+':
                self.turtle.turn(1)
                continue

            if c == '-':
                self.turtle.turn(-1)
                continue

            if c == '[':
                self.turtle.push()
                continue

            if c == ']':
                self.turtle.pop()
                continue

            if c.isupper():
                w, h = self.width, self.height
                self.field[self.turtle.y % h][self.turtle.x % w] \
                    |= (1 << self.turtle.d)
                self.turtle.move()
                self.field[self.turtle.y % h][self.turtle.x % w] \
                    |= (1 << ((self.turtle.d+2) % 4))
                continue

            if c.islower():
                self.turtle.move()

        return self

    def render(self):
        chars = " ─│╰──╯┴│╭│├╮┬┤┼"
        return list(map(lambda l: ''.join(
            map(lambda c: chars[c], l)), self.field))


class LSystem:
    def __init__(self, axiom, rules):
        self.axiom = axiom
        self.rules = rules

    def compute(self, n):
        state = self.axiom
        for i in range(n):
            state = ''.join(map(
                lambda c: self.rules.get(c, c),
                state))
        return state


def make_frame(w, h):
    line = 'F' * (w - 1) + '-' + 'F' * (h - 1) \
        + '-' + 'F' * (w - 1) + '-' + 'F' * (h - 1)
    cx = Canvas(w, h)
    cx.put(line)
    return cx.render()


if __name__ == "__main__":
    with TermApp() as tout:
        width = tout.get_cols()
        height = tout.get_lines()

        cx = Canvas(28, 24)
        cx.turtle.jump(13, 17, 0)
        cx.put(LSystem('F', {'F': 'FF-F-'}).compute(6))
        ls = cx.render()

        fr1 = make_frame(width - 18, 7)
        fr2 = make_frame(width - 18, 12)

        tout << tout.reset_color() << ['setfg', 140, 0, 40]
        tout << ['clear'] << ['cup', 1, 1]
        for l in ls:
            tout << ['sc'] << l << ['rc'] << ['cud', 1]
        tout << ['setfg', 170, 110, 10] << ['cup', 10, 10] << 'λみ'

        tout << ['setfg', 50, 120, 90] << ['cup', 5, 16]
        for l in fr1:
            tout << ['sc'] << l << ['rc'] << ['cud', 1]

        tout << ['setfg', 50, 80, 120] << ['cud', 1]
        for l in fr2:
            tout << ['sc'] << l << ['rc'] << ['cud', 1]

        tout << ['setfg', 160, 220, 100] << ['cup', 3, 20]
        tout << ['sitm']
        tout << " Noodles 0.1.0 - worker console "
        tout << ['cup', 4, 20] << ['rev'] << ['setfg', 40, 40, 40] \
             << "▆"*(width - 24)

        tout << ['sgr0'] << ['op'] << ['setfg', 100, 160, 60] << ['sitm']
        tout << ['cup', 5, 19] << "(status)"
        tout << ['cup', 13, 19] << "(job list)"
        tout << ['cup', 25, 1] << ['op'] << ['sgr0']



from .connection import (Connection)
from .job_keeper import (JobKeeper)

from ..workflow import (
    is_workflow, get_workflow, insert_result,
    is_node_ready, Workflow)
from ..interface import (JobException)
import sys


class Job:
    def __init__(self, workflow, node_id):
        self.workflow = workflow
        self.node_id = node_id

    def __iter__(self):
        return iter((self.workflow, self.node_id))

    @property
    def node(self):
        return self.workflow.nodes[self.node_id]


class DynamicLink:
    def __init__(self, source, target, node):
        self.source = source
        self.target = target
        self.node = node

    def __iter__(self):
        return iter((self.source, self.target, self.node))


error_msg_1 = \
    "A job reported an unexpected error. If this error is not so " \
    "unexpected, consider capturing it with an error handler inside " \
    "the scheduled function or with an error handler in the scheduler " \
    "for a more graceful display and exit.\n" \
    "Infringing job: \n        {}\n"


class Scheduler:
    """
    Schedules jobs, recieves results, then schedules more jobs as they
    become ready to compute. This class communicates with a pool of workers
    by means of coroutines.
    """
    def __init__(self, verbose=False, error_handler=None, job_keeper=None):
        if job_keeper is None:
            self.jobs = JobKeeper()
        else:
            self.jobs = job_keeper
        self.dynamic_links = self.jobs.workflows
        # I'd rather say: self.jobs = job_keeper or {}
        # but Python thruthiness of {} is False
        self.count = 0
        self.key_map = {}
        self.verbose = verbose
        self.handle_error = error_handler

    def run(self, connection: Connection, master: Workflow):
        """Run a workflow.

        :param connection:
            A connection giving a sink to the job-queue and a source yielding
            results.
        :type connection: Connection

        :param master:
            The workflow.
        :type master: Workflow
        """
        # initiate worker slave army and take up reins ...
        source, sink = connection.setup()

        # schedule work
        self.add_workflow(master, master, master.root, sink)
        graceful_exit = False

        # process results
        for job_key, status, result, err_msg in source:
            if status == 'aborted':
                print("Got a fatal error: exiting.",
                      file=sys.stderr, flush=True)
                sys.exit()

            wf, n = self.jobs[job_key]
            if status == 'error':
                if self.handle_error and isinstance(err_msg, JobException):
                    if self.handle_error(wf.nodes[n], *err_msg):
                        graceful_exit = True
                    else:
                        err_msg.reraise()

                else:
                    if isinstance(err_msg, JobException):
                        err_msg.reraise()
                    elif isinstance(err_msg, Exception):
                        raise err_msg
                    else:
                        raise RuntimeError(
                            error_msg_1.format(wf.nodes[n]) + "\n"
                            "Exception raised: {}".format(err_msg))

            if self.verbose:
                print("sched result [{0}]: ".format(self.key_map[job_key]),
                      result,
                      file=sys.stderr, flush=True)

            del self.jobs[job_key]
            if len(self.jobs) == 0 and graceful_exit:
                return

            # if this result is the root of a workflow, pop to parent
            # we do this before scheduling a child workflow, as to
            # achieve tail-call elimination.
            while n == wf.root and wf is not master:
                child = id(wf)
                _, wf, n = self.dynamic_links[child]
                del self.dynamic_links[child]

            # if we retrieve a workflow, push a child
            if is_workflow(result):
                child_wf = get_workflow(result)
                self.add_workflow(child_wf, wf, n, sink)
                continue

            # insert the result in the nodes that need it
            wf.nodes[n].result = result
            for (tgt, address) in wf.links[n]:
                insert_result(wf.nodes[tgt], address, result)
                if is_node_ready(wf.nodes[tgt]) and not graceful_exit:
                    self.schedule(Job(workflow=wf, node_id=tgt), sink)

            # see if we're done
            if wf == master and n == master.root:
                return result

        print("Seventh circle of HELL")

    def schedule(self, job, sink):
        sink.send(self.jobs.register(job))

    def add_workflow(self, wf, target, node, sink):
        self.dynamic_links[id(wf)] = DynamicLink(
            source=wf, target=target, node=node)

        for n in wf.nodes:
            if is_node_ready(wf.nodes[n]):
                self.schedule(Job(workflow=wf, node_id=n), sink)



"""
Manage IO between remote worker/pilot job, and the scheduler. Here there are
two options: use json, or msgpack.
"""

from ..coroutine import coroutine

try:
    import msgpack
except ImportError:
    pass


def MsgPackObjectReader(registry, fi, deref=False):
    yield from msgpack.Unpacker(
        fi, object_hook=lambda o: registry.decode(o, deref),
        encoding='utf-8')


@coroutine
def MsgPackObjectWriter(registry, fo, host=None):
    while True:
        obj = yield
        fo.write(registry.to_msgpack(obj, host=host))
        fo.flush()


def JSONObjectReader(registry, fi, deref=False):
    for line in fi:
        yield registry.from_json(line, deref=deref)


@coroutine
def JSONObjectWriter(registry, fo, host=None):
    # import sys
    while True:
        obj = yield
        # obj_msg = registry.to_json(obj, host=host)
        # print(obj_msg, file=sys.stderr)
        print(registry.to_json(obj, host=host), file=fo, flush=True)



from noodles import schedule, run_parallel, run_process, serial
from math import log, floor


@schedule
def mul(a, b):
    return a * b


@schedule
def factorial(n):
    if n == 0:
        return 1
    else:
        return mul(n, factorial(n - 1))


def test_recursion_parallel():
    f100 = factorial(50.0)
    result = run_parallel(f100, n_threads=1)
    print(result)
    assert floor(log(result)) == 148


def registry():
    return serial.pickle() + serial.base()


def test_recursion_process():
    f100 = factorial(50.0)
    result = run_process(f100, n_processes=1, registry=registry, verbose=False)
    print(result)
    assert floor(log(result)) == 148




class JobException:
    def __init__(self, exc_type, exc_value, exc_tb):
        self.exc_type = exc_type
        self.exc_value = exc_value
        # self.exc_tb = exc_tb

    def reraise(self):
        raise self.exc_value

    def __iter__(self):
        return iter((self.exc_type, self.exc_value, None))

    # def __getstate__(self):
    #     return self.__dict__
    #
    # def __setstate__(self, state):
    #     self.__dict__.update(state)



from collections import namedtuple
from noodles import serial
from noodles.serial.namedtuple import (SerNamedTuple)
from noodles.serial import (Registry)

A = namedtuple('A', ['x', 'y'])
B = namedtuple('B', ['u', 'v'])


def registry1():
    return Registry(
        parent=serial.base(),
        types={
            A: SerNamedTuple(A)
        })


def registry2():
    return serial.base() + serial.namedtuple()


def test_namedtuple_00():
    a = A(73, 38)
    r = registry1()
    assert r.deep_decode(r.deep_encode(a)) == a

    b = B(783, 837)
    r = registry2()
    print(r.deep_encode(b))
    assert r.deep_decode(r.deep_encode(b)) == b



from noodles import (
        schedule, run_single, quote, unquote, run_process,
        schedule_hint, run_logging, find_first)
from noodles.tutorial import (add, sub, mul)
from noodles.serial import base
from noodles.display import NCDisplay


def cond(t, wt, wf):
    return s_cond(t, quote(wt), quote(wf))


@schedule
def s_cond(truth, when_true, when_false):
    if truth:
        return unquote(when_true)
    else:
        return unquote(when_false)


@schedule
def should_not_run():
    raise RuntimeError("This function should never have been called.")


def test_truthfulness():
    w = sub(4, add(3, 2))
    a = cond(True, w, should_not_run())
    b = cond(False, should_not_run(), w)
    result = run_single(mul(a, b))
    assert result == 1


counter = 0


def is_sixteen(n):
    return n == 16


@schedule
def counted_sqr(x):
    global counter
    counter += 1
    return x*x


@schedule_hint(display="squaring {x}", confirm=True)
def display_sqr(x):
    return x*x


def test_find_first():
    global counter

    wfs = [counted_sqr(x) for x in range(10)]
    w = find_first(is_sixteen, wfs)
    result = run_single(w)
    assert result == 16
    assert counter == 5

    wfs = [counted_sqr(x) for x in range(10)]
    w = find_first(is_sixteen, wfs)
    result = run_process(w, n_processes=1, registry=base)
    assert result == 16

    wfs = [display_sqr(x) for x in range(10)]
    w = find_first(is_sixteen, wfs)
    with NCDisplay() as display:
        result = run_logging(w, n_threads=2, display=display)
    assert result == 16



from .registry import (Serialiser, Registry)
from ..utility import look_up
import numpy
import uuid
import io
import base64
import filelock
import h5py
import msgpack
import hashlib


class SerNumpyArray(Serialiser):
    def __init__(self):
        super(SerNumpyArray, self).__init__(numpy.ndarray)

    def encode(self, obj, make_rec):
        fo = io.BytesIO()
        numpy.save(fo, obj, allow_pickle=False)
        return make_rec(base64.b64encode(fo.getvalue()).decode())

    def decode(self, cls, data):
        fi = io.BytesIO(base64.b64decode(data.encode()))
        return numpy.load(fi)


class SerNumpyArrayToFile(Serialiser):
    def __init__(self, file_prefix=None):
        super(SerNumpyArrayToFile, self).__init__(numpy.ndarray)
        self.file_prefix = file_prefix if file_prefix else ''

    def encode(self, obj, make_rec):
        filename = self.file_prefix + str(uuid.uuid4()) + '.npy'
        numpy.save(filename, obj)
        return make_rec(filename, ref=True, files=[filename])

    def decode(self, cls, filename):
        return numpy.load(filename)


def array_sha256(a):
    dtype = msgpack.dumps(str(a.dtype))
    shape = msgpack.dumps(a.shape)
    bdata = a.flatten().view(numpy.uint8)
    sha = hashlib.sha256()
    sha.update(dtype)
    sha.update(shape)
    sha.update(bdata)
    return sha.hexdigest()


class SerNumpyArrayToHDF5(Serialiser):
    def __init__(self, filename, lockfile):
        super(SerNumpyArrayToHDF5, self).__init__(numpy.ndarray)
        self.filename = filename
        self.lock = filelock.FileLock(lockfile)

    def encode(self, obj, make_rec):
        key = array_sha256(obj)
        with self.lock:
            f = h5py.File(self.filename)
            path = next(
                (ds for ds in f if f[ds].attrs.get('hash') == key),
                False)
            if not path:
                path = base64.b64encode(uuid.uuid4().bytes).decode()
                dataset = f.create_dataset(
                    path, shape=obj.shape, dtype=obj.dtype)
                dataset[...] = obj
                dataset.attrs['hash'] = key
                f.close()

        return make_rec({
                "filename": self.filename,
                "path": path
            }, files=[self.filename], ref=True)

    def decode(self, cls, data):
        with self.lock:
            f = h5py.File(self.filename)
            obj = f[data["path"]].value
            f.close()

        return obj


class SerUFunc(Serialiser):
    def __init__(self):
        super(SerUFunc, self).__init__(None)

    def encode(self, obj, make_rec):
        return make_rec('numpy.' + obj.name)

    def decode(self, cls, data):
        return look_up(data)


def _numpy_hook(obj):
    if isinstance(obj, numpy.ufunc):
        return '<ufunc>'

    return None


def arrays_to_file(file_prefix=None):
    """Returns a serialisation registry for serialising NumPy data and
    as well as any UFuncs that have no normal way of retrieving
    qualified names."""
    return Registry(
        types={
            numpy.ndarray: SerNumpyArrayToFile(file_prefix)
        },
        hooks={
            '<ufunc>': SerUFunc()
        },
        hook_fn=_numpy_hook
    )


def arrays_to_string(file_prefix=None):
    return Registry(
        types={
            numpy.ndarray: SerNumpyArray()
        },
        hooks={
            '<ufunc>': SerUFunc()
        },
        hook_fn=_numpy_hook
    )


def arrays_to_hdf5(filename="cache.hdf5"):
    return Registry(
        types={
            numpy.ndarray: SerNumpyArrayToHDF5(filename, "cache.lock")
        },
        hooks={
            '<ufunc>': SerUFunc()
        },
        hook_fn=_numpy_hook
    )

registry = arrays_to_string



from ..queue import Queue
from ..connection import Connection
from ..haploid import (push)

from ..messages import (
    ResultMessage)

from ..remote.io import (
    JSONObjectWriter, JSONObjectReader)

import xenon
import jpype

import threading
import sys
from collections import namedtuple

from .xenon import (XenonKeeper, XenonConfig, XenonScheduler)


def xenon_interactive_worker(XeS: XenonScheduler, job_config):
    """Uses Xenon to run a single remote interactive worker.

    Jobs are read from stdin, and results written to stdout.

    :param XeS:
        The :py:class:`XenonScheduler` object that allows us to schedule the
        new worker.
    :type Xe: XenonScheduler

    :param job_config:
        Job configuration. Specifies the command to be run remotely.
    """
    cmd = job_config.command_line()
    queue = job_config.queue
    J = XeS.submit(cmd, interactive=True, queue=queue)

    def read_stderr():
        jpype.attachThreadToJVM()
        for line in xenon.conversions.read_lines(J.streams.getStderr()):
            print(job_config.name + ": " + line, file=sys.stderr, flush=True)

    J.stderr_thread = threading.Thread(target=read_stderr)
    J.stderr_thread.daemon = True
    J.stderr_thread.start()

    registry = job_config.registry()

    @push
    def send_job():
        output_stream = xenon.conversions.OutputStream(J.streams.getStdin())
        yield from JSONObjectWriter(registry, output_stream, host="scheduler")

    # @pull
    def get_result():
        input_stream = xenon.conversions.InputStream(J.streams.getStdout())
        yield from JSONObjectReader(registry, input_stream)

    return Connection(get_result, send_job, aux=J)


class XenonInteractiveWorker(Connection):
    """Submits a new job to the specified queue. Starts a thread to
    read and forward stderr. Then acts as a `Connection` that goes
    online as soon as the submitted job is running.
    """
    def __init__(self, sched, job_config):
        connection = xenon_interactive_worker(sched, job_config)
        super(XenonInteractiveWorker, self).__init__(*connection)

        self.job_config = job_config
        self.job = connection.aux
        self.online = False

    def init_worker(self):
        """Sends the `init` and `finish` jobs to the worker if this is needed.
        This part will be called first, after a remote worker goes online."""
        pass
        # print("Initializing worker: ", end='', flush=True)
        # if self.job_config.init is not None:
        #     print("[init] ", end='', flush=True)
        #     self.sink().send(
        #         ("init", self.job_config.init()._workflow.root_node))
        #     key, status, result, err_msg = next(self.source())
        #     if key != "init" or not result:
        #         raise RuntimeError(
        #             "The initializer function did not succeed on worker.")
        #
        # if self.job_config.finish is not None:
        #     self.sink().send(
        #         ("finish", self.job_config.finish()._workflow.root_node))
        # print("[done]", flush=True)

    def wait_until_running(self, callback=None):
        """Waits until the remote worker is running, then calls the callback.
        Usually, this method is passed to a different thread; the callback
        is then a function patching results through to the result queue."""
        jpype.attachThreadToJVM()
        status = self.job.wait_until_running(self.job_config.time_out)
        if status.isRunning():
            self.online = True
            self.init_worker()
            if callback:
                callback(self)
        else:
            raise TimeoutError("Timeout while waiting for worker to run: " +
                               self.job_config.name)


RemoteWorker = namedtuple('RemoteWorker', [
    'name', 'lock', 'max', 'jobs', 'source', 'sink'])


class DynamicPool(Connection):
    """Keeps a pool of Xenon workers. At each instant, when a job is sent to
    the recieving coroutine associated with this `Connection`, we look if
    there is a worker idle and send it there. If no worker is idle, the job
    is pushed on the queue. Whenever a worker returns a result, we replennish
    its list of jobs from the queue. In this way we can supply a variable
    amount of workers, each with their own preferred amount of jobs. As of
    yet, we have no mechanism to distribute the jobs between workers in any
    way more knowledgable than that."""
    def __init__(self, Xe: XenonKeeper, xenon_config: XenonConfig):
        self.workers = {}
        self.XeS = XenonScheduler(Xe, xenon_config)
        self.job_queue = Queue()
        self.result_queue = Queue()
        self.lock = threading.Lock()

        @push
        def dispatch_jobs():
            job_sink = self.job_queue.sink()

            while True:
                key, job = yield

                for w in self.workers.values():
                    with w.lock:  # Worker lock ~~~~~~~~~~~~~~~~~~~~~
                        if len(w.jobs) < w.max:
                            w.sink.send((key, job))
                            w.jobs.append(key)
                            break
                    # lock end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                else:
                    job_sink.send((key, job))

        super(DynamicPool, self).__init__(
            self.result_queue.source, dispatch_jobs)

    def add_xenon_worker(self, job_config):
        """Adds a worker to the pool; sets gears in motion."""
        c = XenonInteractiveWorker(self.XeS, job_config)
        w = RemoteWorker(
            job_config.name, threading.Lock(),
            job_config.n_threads, [],
            *c.setup())

        with self.lock:
            self.workers[job_config.name] = w

        def populate(job_source):
            """Populate the worker with jobs, if jobs are available."""
            with w.lock, self.lock:  # Worker lock ~~~~~~~~~~~~~~~~~~
                while len(w.jobs) < w.max and not self.job_queue.empty():
                    key, job = next(job_source)
                    w.sink.send((key, job))
                    w.jobs.append(key)
            # lock end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        def activate(_):
            """Activate the worker."""
            jpype.attachThreadToJVM()

            job_source = self.job_queue.source()
            populate(job_source)

            sink = self.result_queue.sink()

            for result in w.source:
                sink.send(result)

                # do bookkeeping and submit a new job to the worker
                with w.lock:  # Worker lock ~~~~~~~~~~~~~~~~~~~~~
                    w.jobs.remove(result.key)
                populate(job_source)
                # lock end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            for key in w.jobs:
                sink.send(ResultMessage(
                    key, 'aborted', None, 'connection to remote worker lost.'))

        # Start the `activate` function when the worker goes online.
        threading.Thread(
            target=c.wait_until_running, args=(activate,),
            daemon=True).start()



from collections import namedtuple
from ..utility import (unwrap)
from .arguments import (bind_arguments, get_arguments, ref_argument,
                        serialize_arguments, Empty)

NodeData = namedtuple(
    'NodeData',
    ['function', 'arguments', 'hints'])


class FunctionNode:
    """Captures a function call as a combination of function and arguments.
    Some of these arguments may be set to :py:obj:`Empty`, these need to be
    filled in by the workflow before the function can be applied.

    .. py:attribute:: foo

    The function (or object) that is being called.

    .. py:attribute:: bound_args

    A :py:class:`BoundArguments` object storing the arguments to
    the function.
    """
    @staticmethod
    def from_node_data(data):
        foo = unwrap(data.function)
        bound_args = bind_arguments(foo, data.arguments)
        return FunctionNode(foo, bound_args, data.hints)

    def __init__(self, foo, bound_args, hints, result=Empty):
        self.foo = foo
        self.bound_args = bound_args
        self.hints = hints
        self.result = result
        self.prov = None

    def apply(self):
        return self.foo(*self.bound_args.args, **self.bound_args.kwargs)

    @property
    def data(self):
        """Convert to a :py:class:`NodeData` for subsequent serial."""
        return NodeData(
            self.foo, get_arguments(self.bound_args), self.hints)

    def __str__(self):
        s = self.foo.__name__ + '(' + \
            ", ".join(map(str, self.bound_args.args)) + ')'
        if self.result != Empty:
            s += ' -> ' + str(self.result)
        return s


class Workflow:
    """
    The workflow data container.

    .. py:attribute:: root

        A reference to the root node in the graph.

    .. py:attribute:: nodes

        A `dict` listing the nodes in the graph. We use a `dict` only to have
        a persistent object reference.

    .. py:attribute:: links

        A `dict` giving a `set` of links from each node.
    """
    def __init__(self, root, nodes, links):
        self.root = root
        self.nodes = nodes
        self.links = links

    def __iter__(self):
        return iter((self.root, self.nodes, self.links))

    @property
    def root_node(self):
        return self.nodes[self.root]

    @property
    def prov(self):
        return self.root_node.prov

    @property
    def inverse_links(self):
        try:
            return self._inverse_links
        except AttributeError:
            self.create_inverse_links()
            return self._inverse_links

    def create_inverse_links(self):
        self._inverse_links = {}
        for k, v in self.links.items():
            for target, _ in v:
                if target not in self.inverse_links:
                    self._inverse_links[target] = set()
                self._inverse_links[target].add(k)


def is_workflow(x):
    return isinstance(x, Workflow) or ('_workflow' in dir(x))


def get_workflow(x):
    if isinstance(x, Workflow):
        return x

    if '_workflow' in dir(x):
        return x._workflow

    return None


def is_node_ready(node):
    """Returns True if none of the argument holders contain any `Empty` object.
    """
    return all(ref_argument(node.bound_args, a) is not Empty
               for a in serialize_arguments(node.bound_args))



from .registry import (Serialiser, Registry)

import base64
import pickle


class PickleString(Serialiser):
    def __init__(self, cls):
        super(PickleString, self).__init__(cls)

    def encode(self, obj, make_rec):
        data = base64.b64encode(pickle.dumps(obj)).decode('ascii')
        return make_rec(data)

    def decode(self, cls, data):
        return pickle.loads(base64.b64decode(data.encode('ascii')))


def registry():
    """Returns a serialisation registry that "just pickles everything".

    This registry can be used to bolt-on other registries and keep the
    pickle as the default. The objects are first pickled to a byte-array,
    which is subsequently encoded with base64."""
    return Registry(
        default=PickleString(object)
    )



"""
A base class for making objects serializable to JSON.
"""

from copy import deepcopy
from noodles.interface.decorator import PromisedObject, schedule


def storable(obj):
    return isinstance(obj, Storable)


def copy_if_normal(obj, memo):
    if isinstance(obj, PromisedObject):
        return obj
    else:
        return deepcopy(obj, memo)


def from_dict(cls, **kwargs):
    obj = cls.from_dict(**kwargs)
    return obj


class Storable:
    def as_dict(self):
        """Converts the object to a `dict` containing the members
        of the object by name.

        The default implementation is just
        ::

            def as_dict(self):
                return self.__dict__

        In most cases, this method is overloaded to provide a more
        advanced method of serializing data, possibly saving data to
        an external file. In this case the corresponding filename needs
        to be appended to `self.files`.
        """
        d = dict(self.__dict__)
        return d

    @classmethod
    def from_dict(cls, **kwargs):
        """Gets a dictionary by `**kwargs`, and restores the original
        object. For any object `obj` of type `cls` that is derived from
        `Storable`, the following should  be true:
        ::

            cls.from_dict(**obj.as_dict()) == obj

        """
        obj = cls.__new__(cls)
        obj.__dict__ = kwargs
        return obj

    def __deepcopy__(self, memo):
        cls = self.__class__
        tmp = {k: copy_if_normal(v, memo) for k, v in self.as_dict().items()}

        if any(isinstance(x, PromisedObject) for x in tmp.values()):
            return schedule(from_dict)(cls, **tmp)
        else:
            return cls.from_dict(**tmp)



import noodles
from noodles.run.runners import run_single
from noodles.tutorial import add


def test_empty_gather():
    d = noodles.gather()
    result = run_single(d)
    assert result == []


def test_gather():
    d = noodles.gather(*[add(x, x) for x in range(10)])
    result = run_single(d)
    assert result == list(range(0, 20, 2))


def test_gather_dict():
    d = noodles.gather_dict(a=1, b=add(1, 1), c=add(2, 3))
    result = run_single(d)
    assert result == {'a': 1, 'b': 2, 'c': 5}


def test_gather_all():
    d = noodles.gather_all(add(x, 2*x) for x in range(10))
    result = run_single(d)
    assert result == list(range(0, 30, 3))



from importlib import import_module
# import sys
# from copy import copy
# from functools import partial


def object_name(obj):
    return obj.__module__ + '.' + obj.__name__


def look_up(name):
    path = name.split('.')
    module = import_module('.'.join(path[:-1]))
    return getattr(module, path[-1])


def importable(obj):
    try:
        return look_up(object_name(obj)) is obj
    except:
        return False


def unzip_dict(d):
    a = {}
    b = {}

    for k, (v, w) in d.items():
        a[k] = v
        b[k] = w

    return a, b


def unwrap(f):
    try:
        return f.__wrapped__
    except AttributeError:
        return f


class on:
    def __init__(self, env, obj):
        self.env = env
        self.obj = obj

    def __getattr__(self, name):
        return lambda *args: on(self.env, self.env[name](self.obj, *args))


def deep_map(f, root):
    result = f(root)

    if isinstance(result, dict):
        return {k: deep_map(f, v) for k, v in result.items()}

    if isinstance(result, list) or isinstance(result, tuple):
        return [deep_map(f, v) for v in result]

    return result


def inverse_deep_map(f, root):
    if isinstance(root, dict):
        r = {k: inverse_deep_map(f, v) for k, v in root.items()}
    elif isinstance(root, list):
        r = [inverse_deep_map(f, v) for v in root]
    else:
        r = root

    return f(r)



from noodles.interface import (
    delay, gather, lift, schedule, schedule_hint, unwrap,
    has_scheduled_methods, update_hints, unpack, quote, unquote,
    find_first, gather_dict, result, gather_all, maybe)
from noodles.workflow import (get_workflow)
from .run.runners import run_parallel_with_display as run_logging
from .run.runners import (run_single, run_parallel)
from .run.process import run_process
from .run.scheduler import Scheduler
from .storable import Storable

__version__ = "0.2.4"

__all__ = ['schedule', 'schedule_hint', 'run_single', 'run_process',
           'run_logging', 'run_parallel', 'unwrap', 'get_workflow',
           'Scheduler', 'Storable', 'has_scheduled_methods',
           'gather', 'gather_all', 'gather_dict', 'lift', 'unpack', 'maybe',
           'delay', 'update_hints', 'quote', 'unquote', 'find_first', 'result']



from tinydb import (TinyDB, Query)
import hashlib
from threading import Lock
# from ..utility import on
import time
import sys

try:
    import ujson as json
except ImportError:
    import json


def update_object_hash(m, obj):
    r = json.dumps(obj, sort_keys=True)
    m.update(r.encode())
    return m


def prov_key(job_msg, extra=None):
    """Retrieves a MD5 sum from a function call. This takes into account the
    name of the function, the arguments and possibly a version number of the
    function, if that is given in the hints.
    This version can also be auto-generated by generating an MD5 hash from the
    function source. However, the source-code may not always be reachable, or
    the result may depend on an external process which has its own
    versioning."""
    m = hashlib.md5()
    update_object_hash(m, job_msg['data']['function'])
    update_object_hash(m, job_msg['data']['arguments'])

    if 'version' in job_msg['data']['hints']:
        update_object_hash(m, job_msg['data']['hints']['version'])

    if extra is not None:
        update_object_hash(m, extra)

    return m.hexdigest()


def time_stamp():
    return time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime(time.time()))


def attach_job(key):
    def attach_to(rec):
        rec['attached'].append(key)
        return rec
    return attach_to


def append_link(preprov):
    def append_to(rec):
        rec['links'].append(preprov)
        return rec
    return append_to


class JobDB:
    """Keeps a database of jobs, with a MD5 hash that encodes the function
    name, version, and all arguments to the function.
    """
    def __init__(self, path):
        self.db = TinyDB(path)
        self.lock = Lock()

    def get_result_or_attach(self, key, prov, running):
        job = Query()
        with self.lock:
            rec = self.db.get(job.prov == prov)

            if 'result' in rec:
                return 'retrieved', rec['key'], rec['result']

            job_running = rec['key'] in running
            wf_running = rec['link'] in running.workflows

            if job_running or wf_running:
                self.db.update(attach_job(key), job.prov == prov)
                return 'attached', rec['key'], None

            print("WARNING: unfinished job in database. Removing it and "
                  " rerunning.", file=sys.stderr)
            self.db.remove(eids=[rec.eid])
            return 'broken', None, None

    def job_exists(self, prov):
        job = Query()
        with self.lock:
            return self.db.contains(job.prov == prov)

    def store_result(self, key, result):
        job = Query()
        with self.lock:
            if not self.db.contains(job.key == key):
                return

        self.add_time_stamp(key, 'done')
        with self.lock:
            self.db.update(
                    {'result': result, 'link': None},
                    job.key == key)
            rec = self.db.get(job.key == key)
            return rec['attached']

    def new_job(self, key, prov, job_msg):
        with self.lock:
            self.db.insert({
                'key': key,
                'attached': [],
                'prov': prov,
                'link': None,
                'time': {'schedule': time_stamp()},
                'version': job_msg['data']['hints'].get('version'),
                'function': job_msg['data']['function'],
                'arguments': job_msg['data']['arguments']
            })

        return key, prov

    def add_link(self, key, ppn):
        job = Query()
        with self.lock:
            self.db.update({'link': ppn}, job.key == key)

    def get_linked_jobs(self, ppn):
        job = Query()
        with self.lock:
            rec = self.db.search(job.link == ppn)
            return [r['key'] for r in rec]

    def add_time_stamp(self, key, name):
        def update(r):
            r['time'][name] = time_stamp()

        job = Query()
        with self.lock:
            self.db.update(
                update,
                job.key == key)



from ..serial import (Serialiser)


class Path:
    def __init__(self, path):
        self.path = path


class SerPath(Serialiser):
    def __init__(self):
        super(SerPath, self).__init__(Path)

    def encode(self, obj, make_rec):
        return make_rec({'path': obj.path}, files=[obj.path])

    def decode(self, cls, data):
        return cls(path=data['path'])



from .decorator import (
    PromisedObject, schedule, schedule_hint, has_scheduled_methods,
    unwrap, update_hints, result)
from .functions import (
    delay, gather, gather_all, gather_dict, lift, unpack, quote,
    unquote, find_first)
from .annotated_value import (AnnotatedValue)
from .exceptions import (JobException)
from .maybe import (maybe)

__all__ = ['delay', 'gather', 'gather_all', 'gather_dict', 'schedule_hint',
           'schedule', 'unpack', 'has_scheduled_methods', 'unwrap',
           'update_hints', 'AnnotatedValue', 'JobException', 'lift',
           'PromisedObject', 'quote', 'unquote', 'find_first', 'result',
           'maybe']



from noodles import schedule, run_single, run_process, serial, lift
from noodles.storable import Storable
from pytest import raises


class A(Storable):
    def __init__(self):
        super(A, self).__init__()


class M(object):
    pass


@schedule
def f(x, y):
    return x + y


@schedule
def g(a, b):
    return a.x + b.x


@schedule
def h(a, b):
    c = A()
    c.x = a.x - b.x
    return c


def test_storable():
    a = A()
    b = A()

    a.x = 1
    b.x = f(3, 4)

    c = g(a, lift(b))
    b.x = c
    d = h(lift(b), a)

    result = run_process(d, n_processes=1, registry=serial.base)
    assert result.x == 7


def test_nonstorable():
    with raises(TypeError):
        a = A()
        b = M()

        a.x = 1
        b.x = f(3, 4)

        c = g(a, b)
        result = run_single(c)
        assert result == 8



"""
Messages to facilitate communication between scheduler and remote workers.
There are currently three types of messages:

    * ``JobMessage``, sending a job to a worker.
    * ``ResultMessage``, a worker returning a result.
    * ``PilotMessage``, extra communication back and forth, status updates,
    performance information, but also stopping a worker in a nice way.
"""

from ..serial import Reasonable


class JobMessage(Reasonable):
    def __init__(self, key, node):
        self.key = key
        self.node = node

    def __iter__(self):
        return iter((self.key, self.node))
        
        
class ResultMessage(Reasonable):
    def __init__(self, key, status, value, msg):
        self.key = key
        self.status = status
        self.value = value
        self.msg = msg

    def __iter__(self):
        return iter((self.key, self.status, self.value, self.msg))


class PilotMessage(Reasonable):
    def __init__(self, msg, **kwargs):
        self.msg = msg
        self.__dict__.update(kwargs)




import queue
from .haploid import (push, pull)
from .connection import Connection


class Queue(Connection):
    """
    We mock a server/client situation by creating a pipe object that
    recieves items in a sink, stores them in a synchronised queue
    object, and sends them out again in source. Any number of threads
    or objects may create a sink or source. All pool to the same Queue.

    This implementation serves as an example and to glue the local threaded
    runner together. On one side there is a worker pool, taking jobs from one
    of these queues. On the other side there is the controller taking results
    from a second pipe, the snake biting its tail.
    """
    def __init__(self, blocking=True):
        self.Q = queue.Queue()
        self.blocking = blocking

        @push
        def sink():
            while True:
                r = yield
                self.Q.put(r)

        @pull
        def source():
            while True:
                v = self.Q.get()
                yield v
                self.Q.task_done()

        super(Queue, self).__init__(source, sink)

    def empty(self):
        return self.Q.empty()

    def wait(self):
        self.Q.join()



from noodles import schedule, run_single, run_parallel, gather, result
import time


@schedule
def value(a):
    return a


@schedule
def add(a, b):
    return a+b


@schedule
def sub(a, b):
    return a-b


@schedule
def mul(a, b):
    return a*b


@schedule
def sum(a, buildin_sum=sum):
    return buildin_sum(a)


def test_runner_01():
    A = value(1)
    B = value(1)
    C = add(A, B)

    assert run_single(C) == 2
    assert result(A) == 1
    assert result(B) == 1


def test_runner_02():
    A = add(1, 1)
    B = sub(3, A)

    multiples = [mul(add(i, B), A) for i in range(6)]
    C = sum(gather(*multiples))

    assert run_single(C) == 42
    assert result(C) == 42
    assert result(A) == 2


def test_parallel_runner_01():
    A = add(1, 1)
    B = sub(3, A)

    multiples = [mul(add(i, B), A) for i in range(6)]
    C = sum(gather(*multiples))

    assert run_parallel(C, 4) == 42


@schedule
def delayed(a, delay=1.0):
    time.sleep(delay)
    return a


def test_parallel_runner_02():
    A = [delayed(1, 0.01) for i in range(100)]
    B = sum(gather(*A))

    start = time.time()
    assert run_parallel(B, 4) == 100
    end = time.time()
    # print(end-start)
    assert (end - start) < 0.4   # liberal upper limit for running time



def find_links_to(links, node):
    """
    Find links to a node.

    :param links:
        forward links of a workflow
    :type links: Mapping[NodeId, Set[(NodeId, ArgumentType, [int|str]])]

    :param node:
        index to a node
    :type node: int

    :returns:
        dictionary of sources for each argument
    :rtype: Mapping[(ArgumentType, [int|str]), NodeId]
    """
    return {address: src
            for src, (tgt, address) in _all_valid(links)
            if tgt == node}


def _all_valid(links):
    """
    Iterates over all links, forgetting emtpy registers.
    """
    for k, v in links.items():
        for i in v:
            yield k, i


def invert_links(links):
    """
    Inverts the call-graph to get a dependency graph. Possibly slow,
    short version.

    :param links:
        forward links of a call-graph.
    :type links: Mapping[NodeId, Set[(NodeId, ArgumentType, [int|str]])]

    :returns:
        inverted graph, giving dependency of jobs.
    :rtype: Mapping[NodeId, Mapping[(ArgumentType, [int|str]), NodeId]]
    """
    return {node: find_links_to(links, node) for node in links}



from .haploid import (pull_map)
from .messages import (ResultMessage)
from ..interface import (AnnotatedValue, JobException)
from ..utility import (object_name)
import sys


@pull_map
def worker(key, job):
    return run_job(key, job)


def run_job(key, job):
    """A worker coroutine. Pulls jobs, yielding results. If an
    exception is raised running the job, it returns a result
    object with 'error' status. If the job requests return-value
    annotation, a two-tuple is expected; this tuple is then
    unpacked, the first being the result, the second part is
    sent on in the error message slot."""
    try:
        if job.hints and 'annotated' in job.hints:
            result = job.apply()
            if isinstance(result, tuple) and len(result) == 2:
                value, msg = result
                return ResultMessage(key, 'done', value, msg)
            else:
                raise TypeError("You promised annotation in call "
                                "to function {} but return value "
                                "is incompatible with 2-tuple."
                                .format(object_name(job.foo)))

        else:
            result = job.apply()
            if isinstance(result, AnnotatedValue):
                value, message = result
                return ResultMessage(key, 'done', value, message)

            return ResultMessage(key, 'done', result, None)

    except Exception:
        exc_info = sys.exc_info()
        return ResultMessage(key, 'error', None, JobException(*exc_info))



from inspect import Parameter
from graphviz import Digraph

from . import schedule, schedule_hint, get_workflow


# scheduled functions

@schedule
def add(x, y):
    return x + y


@schedule_hint(display="{a} + {b}", confirm=True)
def log_add(a, b):
    return a + b


@schedule
def sub(x, y):
    return x - y


@schedule
def mul(x, y):
    return x*y


@schedule
def accumulate(lst, start=0):
    return sum(lst, start)


# functions for printing workflow graphs in notebooks

def _format_arg_list(a, v):
    if len(a) == 0:
        if v:
            return "(\u2026)"
        else:
            return "()"

    s = "("
    for i in a[:-1]:
        s += str(i) if i != Parameter.empty else "\u2014"
        s += ", "

    if v:
        s += "\u2014"
    else:
        s += str(a[-1]) if a[-1] != Parameter.empty else "\u2014"

    s += ")"
    return s


def get_workflow_graph(promise):
    workflow = get_workflow(promise)

    dot = Digraph()
    for i,n in workflow.nodes.items():
        dot.node(str(i), label="{0} \n {1}".format(
            n.foo.__name__,
            _format_arg_list(n.bound_args.args, None)))

    for i in workflow.links:
        for j in workflow.links[i]:
            dot.edge(str(i), str(j[0]), label=str(j[1].name))

    return dot



"""
This is the Noodles executable worker module. This is usually run by the
scheduler to dispatch jobs remotely, using

.. code-block:: bash

    > python3.5 -m noodles.worker <args>...

There are several modes in which the worker generates results. For all
of these modes, the rule is **one object per line**.

:batch-mode: Give a single job as a JSON string on the command line. The
    result is printed to stdout, again as a JSON object.

    .. code-block:: bash

        > python3.5 -m noodles.worker -batch "{'job': ...}"

:online-mode: Recieve worker commands as JSON objects through stdin and
    send out results to stdout.

    .. code-block:: bash

        > python3.5 -m noodles.worker -online [-use <worker>]
"""

import argparse
import sys
import uuid
from contextlib import redirect_stdout

import os
from .utility import (look_up)

try:
    import msgpack
    has_msgpack = True
except ImportError:
    has_msgpack = False

from .run.worker import (
    run_job)

from .run.messages import (
    JobMessage)

from .run.remote.io import (
    MsgPackObjectReader, MsgPackObjectWriter,
    JSONObjectReader, JSONObjectWriter)


def run_batch_mode(args):
    print("Batch mode is not yet implemented")


def run_online_mode(args):
    if args.n == 1:
        registry = look_up(args.registry)()
        finish = None

        if args.msgpack:
            newin = os.fdopen(sys.stdin.fileno(), 'rb', buffering=0)
            input_stream = MsgPackObjectReader(
                registry, newin, deref=True)
            output_stream = MsgPackObjectWriter(
                registry, sys.stdout.buffer, host=args.name)
        else:
            input_stream = JSONObjectReader(
                registry, sys.stdin, deref=True)
            output_stream = JSONObjectWriter(
                registry, sys.stdout, host=args.name)

        # run the init function if it is given
        if args.init:
            with redirect_stdout(sys.stderr):
                look_up(args.init)()

        if args.finish:
            finish = look_up(args.finish)

        for msg in input_stream:
            if isinstance(msg, JobMessage):
                key, job = msg
            elif isinstance(msg, tuple):
                key, job = msg
            else:
                continue

            if args.jobdirs:
                # make a directory
                os.mkdir("noodles-{0}".format(key.hex))
                # enter it
                os.chdir("noodles-{0}".format(key.hex))

            if args.verbose:
                print("worker: ",
                      job.foo.__name__,
                      job.bound_args.args,
                      job.bound_args.kwargs,
                      file=sys.stderr, flush=True)

            with redirect_stdout(sys.stderr):
                result = run_job(key, job)

            if args.verbose:
                print("result: ", result, file=sys.stderr, flush=True)

            if args.jobdirs:
                # parent directory
                os.chdir("..")

            output_stream.send(result)

        if finish:
            finish()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="{py} -m noodles.worker".format(
            py=os.path.basename(sys.executable)),

        description="This is the Noodles executable worker module. "
                    "This is usually run by the scheduler to dispatch "
                    "jobs remotely.")

    parser.add_argument(
        "-version", action="version", version="Noodles 0.1.0-alpha1")

    subparsers = parser.add_subparsers(
        title="Execution models",
        description="Noodles can execute jobs one at a time, or "
                    "streaming on stdin/stdout.")

    batch_parser = subparsers.add_parser(
        "batch", help="run a single job")
    batch_parser.add_argument(
        "-persist",
        help="keep going until a value, not another workflow, comes out.",
        default=False, action='store_true')
    batch_parser.set_defaults(func=run_batch_mode)

    online_parser = subparsers.add_parser(
        "online", help="stream jobs from standard input.")
    online_parser.add_argument(
        "-registry", type=str,
        help="the serial registry")
    online_parser.add_argument(
        "-n", type=int,
        help="the number of threads.", default=1)
    online_parser.add_argument(
        "-verbose",
        help="output information to stderr for debugging",
        default=False, action='store_true')
    online_parser.add_argument(
        "-jobdirs",
        help="create a directory for each job to run in",
        default=False, action='store_true')
    online_parser.add_argument(
        "-msgpack",
        help="use MessagePack for serialisation.",
        default=False, action='store_true')
    online_parser.add_argument(
        "-name", type=str,
        help="worker identity",
        default="worker-" + str(uuid.uuid4()))
    online_parser.add_argument(
        "-init", type=str,
        help="an init function will be send before other jobs",
        default=None)
    online_parser.add_argument(
        "-finish", type=str,
        help="a finish function will be send before other jobs",
        default=None)

    online_parser.set_defaults(func=run_online_mode)

    args = parser.parse_args()
    args.func(args)



from functools import wraps
from .messages import (ResultMessage)
from .haploid import (push_map, pull_map)
import threading


def source_branch(*flst):
    @pull_map
    def g(*args):
        for f in flst:
            f(*args)
        return args

    return g


def sink_branch(*flst):
    @push_map
    def g(*args):
        for f in flst:
            f(*args)
        return args

    return g


class CatchExceptions(object):
    """Catches exceptions running a function and passes them
    on to a Queue. This should be used when running a critical
    thread.

    The CatchExceptions object has two important members:
    `job_pass` and `result_pass`, the first is a `pull_map`,
    the second a `send_map`, assuming we pull and send to an
    intermediate queue."""
    def __init__(self, sink):
        self.sink = sink
        self.jobs = set()
        self.lock = threading.Lock()

        self.job_sink = sink_branch(self.job_pass)
        self.job_source = source_branch(self.job_pass)
        self.result_sink = sink_branch(self.result_pass)
        self.result_source = source_branch(self.result_pass)

    def job_pass(self, key, job):
        with self.lock:
            self.jobs.add(key)

    def result_pass(self, key, status, result, err):
        with self.lock:
            self.jobs.remove(key)

    def __call__(self, fn):
        @wraps(fn)
        def protected_fn(*args, **kwargs):
            sink = self.sink()
            try:
                return fn(*args, **kwargs)
            except Exception as exc:
                with self.lock:
                    for key in self.jobs:
                        sink.send(ResultMessage(key, 'aborted', None, exc))

                raise exc

        return protected_fn



from .registry import (Serialiser)


class AsDict(Serialiser):
    def __init__(self, cls):
        super(AsDict, self).__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(obj.__dict__)

    def decode(self, cls, data):
        obj = cls.__new__(cls)
        obj.__dict__ = data
        return obj



from noodles import (schedule, Scheduler, has_scheduled_methods, serial)
from noodles.serial import (Registry, AsDict)
from noodles.workflow import get_workflow
from noodles.run.process import process_worker


def registry():
    reg = Registry(parent=serial.base())
    reg[A] = AsDict(A)
    return reg


@has_scheduled_methods
class A:
    def __init__(self, a):
        super(A, self).__init__()
        self.a = a

    @schedule
    def mul(self, b):
        return self.a * b


def test_sched_meth():
    a = A(5)
    b = a.mul(5)
    result = Scheduler().run(process_worker(registry), get_workflow(b))
    assert result == 25



from functools import (wraps)
from itertools import (chain)
import inspect
from ..utility import (object_name)


class Fail:
    def __init__(self, func, fails=None, exception=None):
        self.name = "{} ({}:{})".format(
            object_name(func),
            inspect.getsourcefile(func),
            inspect.getsourcelines(func)[1])
        self.fails = fails or []
        self.trace = []
        self.exception = exception

    def add_call(self, func):
        self.trace.append("{} ({}:{})".format(
            object_name(func),
            inspect.getsourcefile(func),
            inspect.getsourcelines(func)[1]))

        return self

    @property
    def is_root_cause(self):
        return self.exception is not None

    def __bool__(self):
        return False

    def __str__(self):
        msg = "Fail: " + " -> ".join(self.trace + [self.name])
        if self.exception is not None:
            msg += "\n* exception: "
            msg += "\n    ".join(l for l in str(self.exception).split('\n'))
        elif self.fails:
            msg += "\n* failed arguments:\n    "
            msg += "\n    ".join(
                    "{} `{}` ".format(foo, bar) + "\n    ".join(
                        l for l in str(fail).split('\n'))
                    for foo, bar, fail in self.fails)
        return msg


def maybe(f):
    """Calls `f` in a try/except block, returning a `Fail` object if
    the call fails in any way. If any of the arguments to the call are Fail
    objects, the call is not attempted."""

    name = object_name(f)

    @wraps(f)
    def maybe_wrapped(*args, **kwargs):
        fails = [(name, k, v)
            for k, v in chain(enumerate(args), kwargs.items())
            if isinstance(v, Fail)]

        if fails:
            return Fail(f, fails=fails)

        try:
            result = f(*args, **kwargs)

        except Exception as e:
            return Fail(f, exception=e)

        else:
            if isinstance(result, Fail):
                result.add_call(f)

            return result

    return maybe_wrapped



from noodles import run_process, schedule, serial

class A(dict):
    pass

@schedule
def f(a):
    a['value'] = 5
    return a

def test_dict_like():
    a = A()
    b = f(a)
    result = run_process(b, 1, serial.base)

    assert isinstance(result, A)
    assert result['value'] == 5



from noodles import (serial, files)


def test_path():
    a = files.Path("/usr/bin/python")
    registry = serial.base() + files.registry()

    encoded = registry.deep_encode(a)
    b = registry.deep_decode(encoded)

    assert a.path == b.path



import noodles
from noodles.tutorial import (add, sub, mul, accumulate)
from noodles.display import (NCDisplay)
from noodles.run.runners import (
    run_single, run_parallel, run_parallel_with_display)


def test_broker_01():
    A = add(1, 1)
    B = sub(3, A)

    multiples = [mul(add(i, B), A) for i in range(6)]
    C = accumulate(noodles.gather(*multiples))

    assert run_parallel(C, 4) == 42


def test_broker_02():
    A = add(1, 1)
    B = sub(3, A)

    multiples = [mul(add(i, B), A) for i in range(6)]
    C = accumulate(noodles.gather(*multiples))

    assert run_single(C) == 42


@noodles.schedule_hint(display="{a} + {b}", confirm=True)
def log_add(a, b):
    return a + b


@noodles.schedule_hint(display="{msg}")
def message(msg, value=0):
    return value()


def test_broker_logging():
    A = log_add(1, 1)
    B = sub(3, A)

    multiples = [mul(log_add(i, B), A) for i in range(6)]
    C = accumulate(noodles.gather(*multiples))

    with NCDisplay(title="Running the test") as display:
        assert run_parallel_with_display(C, 4, display) == 42



from .haploid import (pull, patch)
from .queue import (Queue)
from .protect import (CatchExceptions)
import threading


def thread_pool(*workers, results=None):
    """Threadpool should run functions. That means that both input and output
    need to be active mode, that this cannot be represented by a simple haploid
    co-routine.

    The resulting object should be able to replace a single worker in the
    chain, looking like::

        @haploid('pull')
        def worker(source):
            for job in source:
                yield run(job)

    To mend this, we run the required number of threads with `patch`, taking
    the workers as input source and a Queue sink as output. Then we yield from
    the Queue's source. The source that this is called with should then be
    thread-safe.
    """
    results = results if results is not None else Queue()

    @pull
    def fn(source):
        for s in workers:
            catch = CatchExceptions(results.sink)

            t = threading.Thread(
                target=catch(patch),
                args=(
                    lambda: (catch.job_source >> s)(source),
                    lambda: catch.result_sink(results.sink)),
                daemon=True)
            t.start()

        yield from results.source()

    return fn



from .dynamic_pool import DynamicPool
from ..scheduler import Scheduler
from ..job_keeper import JobKeeper
from ..haploid import (broadcast, sink_map, patch, branch)
from ..queue import Queue

from ...workflow import (get_workflow)

import threading
from copy import copy

# Only define run_xenon_prov if provenance is installed
try:
    from ..run_with_prov import prov_wrap_connection
except ImportError:
    pass
else:
    def run_xenon_prov(wf, Xe, jobdb_file, n_processes, xenon_config,
                       job_config, *, deref=False, job_keeper=None,
                       display=None):
        """Run the workflow using a number of online Xenon workers.

        :param Xe:
            The XenonKeeper instance.

        :param wf:
            The workflow.
        :type wf: `Workflow` or `PromisedObject`

        :param n_processes:
            Number of processes to start.

        :param xenon_config:
            The :py:class:`XenonConfig` object that gives tells us how to use
            Xenon.

        :param job_config:
            The :py:class:`RemoteJobConfig` object that specifies the command
            to run remotely for each worker.

        :param deref:
            Set this to True to pass the result through one more encoding and
            decoding step with object derefencing turned on.
        :type deref: bool

        :returns: the result of evaluating the workflow
        :rtype: any
        """
        DP = DynamicPool(Xe, xenon_config)

        for i in range(n_processes):
            cfg = copy(job_config)
            cfg.name = 'xenon-{0:02}'.format(i)
            DP.add_xenon_worker(cfg)

        if job_keeper is None:
            job_keeper = JobKeeper()
        S = Scheduler(job_keeper=job_keeper)

        LogQ = Queue()
        if display:
            tgt = broadcast(job_keeper.message, sink_map(display))
        else:
            tgt = job_keeper.message

        threading.Thread(
            target=patch,
            args=(LogQ.source, tgt),
            daemon=True).start()

        result = S.run(
            prov_wrap_connection(
                DP >> branch(LogQ.sink),
                DP.result_queue, job_config.registry,
                jobdb_file, job_keeper,
                log_q=LogQ),
            get_workflow(wf))

        if deref:
            return job_config.registry().dereference(result, host='localhost')
        else:
            return result


def run_xenon(wf, Xe, jobdb_file, n_processes, xenon_config,
              job_config, *, deref=False, job_keeper=None):
    """Run the workflow using a number of online Xenon workers.

    :param Xe:
        The XenonKeeper instance.

    :param wf:
        The workflow.
    :type wf: `Workflow` or `PromisedObject`

    :param n_processes:
        Number of processes to start.

    :param xenon_config:
        The :py:class:`XenonConfig` object that gives tells us how to use
        Xenon.

    :param job_config:
        The :py:class:`RemoteJobConfig` object that specifies the command to
        run remotely for each worker.

    :param deref:
        Set this to True to pass the result through one more encoding and
        decoding step with object derefencing turned on.
    :type deref: bool

    :returns: the result of evaluating the workflow
    :rtype: any
    """

    DP = DynamicPool(Xe, xenon_config)

    for i in range(n_processes):
        cfg = copy(job_config)
        cfg.name = 'xenon-{0:02}'.format(i)
        DP.add_xenon_worker(cfg)

    if job_keeper is None:
        job_keeper = JobKeeper()
    S = Scheduler(job_keeper=job_keeper)

    # results = Queue()

    # LogQ = Queue()
    # if display:
    #     tgt = broadcast(job_keeper.message, sink_map(display))
    # else:
    #     tgt = job_keeper.message
    #
    # threading.Thread(
    #     target=patch,
    #     args=(LogQ.source, tgt),
    #     daemon=True).start()

    result = S.run(
        DP, get_workflow(wf)
    )
    # result = S.run(
    #     prov_wrap_connection(
    #         DP >> branch(LogQ.sink),
    #         results, job_config.registry,
    #         jobdb_file, job_keeper,
    #         log_q=LogQ),
    #     get_workflow(wf))

    if deref:
        return job_config.registry().dereference(result, host='localhost')
    else:
        return result



from .arguments import set_argument, Empty


def reset_workflow(workflow):
    for tgt in workflow.links.values():
        for m, a in tgt:
            set_argument(workflow.nodes[m].bound_args, a, Empty)

    return workflow


def insert_result(node, address, value):
    """Runs `set_argument`, but checks first wether the data location is not
    already filled with some data. In any normal circumstance this checking
    is redundant, but if we don't give an error here the program would continue
    with unexpected results.
    """
    # a = ref_argument(node.bound_args, address)
    # if a != Empty:
    #    raise RuntimeError(
    #        "Noodle panic. Argument {arg} in {name} already given." \
    #        .format(arg=format_address(address),
    #                name=node.foo.__name__))

    set_argument(node.bound_args, address, value)



from noodles import schedule, run_single, unwrap, run_process, Storable
from noodles import serial
from pytest import raises


@schedule
def sqr(x):
    return x*x


@schedule
def divide(x, y):
    return x/y


@schedule
def mul(x, y):
    return x*y


@schedule
def make_object(cls, *args, **kwargs):
    return cls(*args, **kwargs)


class A(Storable):
    def __init__(self, value):
        # super(A, self).__init__()
        self.value = value
        self.m_attr = 0

    @property
    def attr(self):
        return sqr(self.m_attr)

    def mul_attr(self, factor=1):
        return mul(self.m_attr, factor)

    @attr.setter
    def attr(self, x):
        self.m_attr = divide(x, 2)

    def multiply(self, factor):
        self.value *= factor
        return self


class B:
    pass


def test_class_decorator():
    a = make_object(A, 5).multiply(10)
    a.second = 7
    result = run_single(a)
    assert result.value == 50
    assert result.second == 7


def registry():
    return serial.base()


def test_class_decorator_process():
    a = make_object(A, 5).multiply(10)
    a.second = 7
    result = run_process(
        a, n_processes=1, registry=registry)
    assert result.value == 50
    assert result.second == 7


def test_class_property():
    a = make_object(A, 10)
    a.attr = 1.0

    b = make_object(B)
    b.first = a.attr
    b.second = a.mul_attr(3)

    result = run_single(b)
    assert result.first == 0.25
    assert result.second == 1.5

# @schedule
# def should_not_run():
#     raise RuntimeError
#
# def test_setget_optimisation1():
#     a = B()
#     a.r = 5
#     a.s = should_not_run()
#
#     b = B()
#     b.t = a.r
#
#     result = run(b)
#     assert result.t == 5
#
# @raises(RuntimeError)
# def test_setget_optimisation2():
#     a = B()
#     a.s = should_not_run()
#     a.r = 5
#
#     b = B()
#     b.t = a.r
#
#     result = run(b)
#     assert result.t == 5


def f(x):
    return x


def test_unwrap():
    assert f == unwrap(schedule(f))
    assert f == unwrap(f)


def test_class_decorator2():
    with raises(AttributeError):
        a = make_object(A, 6).multiply(7)
        b = a.divide(0)
        result = run_single(b)
        print(dir(result))



import pytest

try:
    from noodles.prov import prov_key
    from noodles.run.run_with_prov import (
        run_single, run_parallel, run_parallel_opt)
except ImportError:
    has_prov = False
else:
    has_prov = True

from noodles import serial, gather, schedule_hint
from noodles.tutorial import (add, mul, sub, accumulate)

from noodles.run.worker import run_job
from noodles.run.job_keeper import JobKeeper

import os


@schedule_hint(store=True)
def add2(x, y):
    return x + y


@schedule_hint(store=True)
def fib(n):
    if n < 2:
        return 1
    else:
        return add(fib(n - 1), fib(n - 2))


@pytest.mark.skipif(not has_prov, reason="TinyDB needed")
class TestProv:
    def test_prov_00(self):
        reg = serial.base()
        a = add(3, 4)
        b = sub(3, 4)
        c = add(3, 4)
        d = add(4, 3)

        enc = [reg.deep_encode(x._workflow.root_node) for x in [a, b, c, d]]
        key = [prov_key(o) for o in enc]
        assert key[0] == key[2]
        assert key[1] != key[0]
        assert key[3] != key[0]

    def test_prov_01(self):
        reg = serial.base()
        a = add(3, 4)

        enc = reg.deep_encode(a._workflow.root_node)
        dec = reg.deep_decode(enc)

        result = run_job(0, dec)
        assert result.value == 7

    def test_prov_02(self):
        db_file = "prov1.json"

        A = add(1, 1)
        B = sub(3, A)

        multiples = [mul(add(i, B), A) for i in range(6)]
        C = accumulate(gather(*multiples))

        result = run_single(C, serial.base, db_file)
        assert result == 42
        os.unlink(db_file)

    def test_prov_03(self):
        db_file = "prov2.json"

        A = add(1, 1)
        B = sub(3, A)

        multiples = [mul(add(i, B), A) for i in range(6)]
        C = accumulate(gather(*multiples))

        result = run_parallel(C, 4, serial.base, db_file, JobKeeper(keep=True))
        assert result == 42
        os.unlink(db_file)

    def test_prov_04(self):
        db_file = "prov3.json"

        A = add2(1, 1)
        B = sub(3, A)

        multiples = [mul(add2(i, B), A) for i in range(6)]
        C = accumulate(gather(*multiples))

        result = run_parallel_opt(C, 4, serial.base, db_file)
        assert result == 42
        os.unlink(db_file)

    def test_prov_05(self):
        import time
        db_file = "testjobs.json"

        wf = fib(20)
        start = time.time()
        result = run_parallel(wf, 4, serial.base, db_file)
        end = time.time()

        assert (end - start) < 5.0  # weak test
        assert result == 10946

        os.unlink(db_file)



from noodles import (schedule, run_single, gather, lift)
from noodles.tutorial import (add, sub, mul)
from collections import OrderedDict


@schedule
def g(x):
    return x['a'] + x['b']

def test_lift_ordered_dict():
    x = OrderedDict()
    x['a'] = 1
    x['b'] = add(1, 2)
    y = g(lift(x))
    result = run_single(y)
    assert result == 4


class A:
    pass


@schedule
def f(a):
    return a.x + a.y


def test_lift_01():
    a = A()
    a.x = add(1, 2)
    a.y = sub(9, 11)
    b = f(lift(a))

    result = run_single(b)
    assert result == 1




# from .numpy import registry as numpy
from .pickle import registry as pickle
from .base import registry as base
from .as_dict import AsDict
from .namedtuple import registry as namedtuple
from .registry import (Registry, Serialiser, RefObject)
from .reasonable import (Reasonable)

__all__ = ['pickle', 'base', 'Registry', 'Serialiser',
           'RefObject', 'AsDict', 'namedtuple', 'Reasonable']



from noodles import schedule, run_single, unpack
from noodles.tutorial import add


@schedule
def f():
    return 1, 2, 3


def test_unpack_00():
    a, b, c = unpack(f(), 3)
    d = run_single(add(a, add(b, c)))
    assert d == 6




from noodles.utility import deep_map


class A:
    def __init__(self, value):
        self.value = value


class B(A):
    pass


def translate(a):
    if isinstance(a, A):
        name = type(a).__name__
        return {'type': name, 'value': a.value}

    else:
        return a


def test_deep_map():
    a = A(5)
    b = A("Hello")
    c = B(67.372)
    d = B([a, b])
    e = A({'one': c, 'two': d})

    result = deep_map(translate, e)
    print(result)
    assert result['value']['two']['value'][0]['value'] == 5
    assert result['value']['one']['type'] == 'B'
    assert result['value']['two']['value'][1]['value'] == "Hello"




import noodles
from noodles.run.runners import (run_parallel_timing)
from noodles.tutorial import (mul, sub, accumulate)
import time
import sys

@noodles.schedule_hint(
    display="adding {a} + {b}")
def add(a, b):
    time.sleep(0.01)
    return a + b

def test_xenon_42():
    A = add(1, 1)
    B = sub(3, A)

    multiples = [mul(add(i, B), A) for i in range(6)]
    C = accumulate(noodles.gather(*multiples))

    result = run_parallel_timing(C, 4, sys.stdout)
    assert(result == 42)




#!/usr/bin/env python

from setuptools import setup
from os import path
from codecs import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='Noodles',
    version='0.2.4',
    description='Workflow Engine',
    author='Johan Hidding',
    url='https://github.com/NLeSC/noodles',
    packages=[
        'noodles', 'noodles.serial', 'noodles.run', 'noodles.run.xenon',
        'noodles.run.remote',
        'noodles.display',
        'noodles.interface', 'noodles.workflow', 'noodles.files',
        'noodles.prov', 'noodles.draw_workflow'],

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Environment :: Console',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.5',
        'Topic :: System :: Distributed Computing'],

    install_requires=['graphviz'],
    extras_require={
        'doc': ['sphinx', 'shpinx_rtd_theme', 'nbsphinx'],
        'prov': ['tinydb', 'ujson'],
        'xenon': ['pyxenon'],
        'numpy': ['numpy', 'h5py', 'msgpack-python', 'filelock'],
        'test': ['pytest', 'pytest-cov', 'codacy-coverage', 'pyflakes', 'pep8',
                 'docker-py']
    },
)



"""
Next generation class of Workflow Runners. These are accessed by configuring
Noodles in a `noodles.ini` file. For this reason they have to obey similar
interfaces.
"""

from .queue import (Queue)
from .scheduler import (Scheduler)
from .haploid import (push_map, sink_map, branch, patch)
from .thread_pool import (thread_pool)
from .worker import (worker)
from ..workflow import (get_workflow)

from itertools import (repeat)
import threading


# =========================================================================== #
def run_single(wf):
    """Run a workflow in a single thread. This is the absolute minimal
    runner, consisting of a single queue for jobs and a worker running
    jobs every time a result is pulled."""
    S = Scheduler()
    W = Queue() >> worker

    return S.run(W, get_workflow(wf))


def run_single_with_display(wf, display):
    """Adds a display to the single runner. Everything still runs in a single
    thread. Every time a job is pulled by the worker, a message goes to the
    display routine; when the job is finished the result is sent to the display
    routine."""

    @push_map
    def log_job_start(key, job):
        return (key, 'start', job, None)

    S = Scheduler(error_handler=display.error_handler)
    W = Queue() \
        >> branch(log_job_start.to(sink_map(display))) \
        >> worker \
        >> branch(sink_map(display))

    return S.run(W, get_workflow(wf))


def run_parallel(wf, n_threads):
    """Run a workflow in `n_threads` parallel threads. Now we replaced the
    single worker with a thread-pool of workers."""
    S = Scheduler()
    W = Queue() >> thread_pool(*repeat(worker, n_threads))

    return S.run(W, get_workflow(wf))


def run_parallel_with_display(wf, n_threads, display):
    """Adds a display to the parallel runner. Because messages come in
    asynchronously now, we start an extra thread just for the display
    routine."""

    @push_map
    def log_job_start(key, job):
        return (key, 'start', job, None)

    LogQ = Queue()

    S = Scheduler(error_handler=display.error_handler)

    threading.Thread(
        target=patch,
        args=(LogQ.source, sink_map(display)),
        daemon=True).start()

    W = Queue() \
        >> branch(log_job_start >> LogQ.sink) \
        >> thread_pool(*repeat(worker, n_threads)) \
        >> branch(LogQ.sink)

    result = S.run(W, get_workflow(wf))
    LogQ.wait()

    return result



import noodles
from pytest import raises


def test_iterate():
    with raises(TypeError):
        a = noodles.delay((1, 2, 3))
        b, c, d = a
        e = noodles.gather(d, c, b)

        result = noodles.run_single(e)
        assert result == (3, 2, 1)



from noodles import schedule, run_parallel, run_single, gather, run_logging
import sys


@schedule
def sqr(a):
    return a*a


@schedule
def sum(a, buildin_sum=sum):
    return buildin_sum(a)


@schedule
def map(f, lst):
    return gather(*[f(x) for x in lst])


@schedule
def num_range(a, b):
    return range(a, b)


def test_higher_order():
    w = sum(map(sqr, num_range(0, 10)))
    assert run_parallel(w, 4) == 285


@schedule
def g(x):
    return f(x)


@schedule
def f(x):
    return x


class Display:
    def __call__(self, key, status, data, err_msg):
        print(status, data, err_msg)

    def error_handler(self, job, xcptn):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


def test_single_node():
    with Display() as display:
        assert run_logging(g(5), 1, display) == 5



from .runner import (run_xenon)
from .xenon import (XenonJob, XenonScheduler, XenonConfig, RemoteJobConfig,
                    XenonKeeper)

# Only export run_xenon_prov if provenance is installed
try:
    from .runner import run_xenon_prov
    __all__ = ['run_xenon_prov']
except ImportError:
    __all__ = []

__all__ += ['XenonConfig', 'XenonJob', 'XenonScheduler', 'RemoteJobConfig',
            'XenonKeeper', 'run_xenon']



from .connection import (Connection)
from .queue import (Queue)
from .scheduler import (Scheduler)
from .messages import (ResultMessage)
from .haploid import (
    pull, push, push_map, sink_map,
    branch, patch, broadcast)
from .thread_pool import (thread_pool)
from .worker import (worker, run_job)
from .job_keeper import (JobKeeper)

from ..workflow import (get_workflow, is_workflow)
from ..prov import (JobDB, prov_key)

from itertools import (repeat)
import threading


def run_single(wf, registry, jobdb_file, display=None):
    """Run a workflow in a single thread. This is the absolute minimal
    runner, consisting of a single queue for jobs and a worker running
    jobs every time a result is pulled.

    This version integrates with the JobDB.
    """
    registry = registry()
    db = JobDB(jobdb_file)

    def decode_result(key, obj):
        return ResultMessage(key, 'retrieved', registry.deep_decode(obj), None)

    @pull
    def pass_job(source):
        for key, job in source():
            job_msg = registry.deep_encode(job)
            prov = prov_key(job_msg)

            if db.job_exists(prov):
                status, _, result = db.get_result_or_attach(
                    key, prov, {})
                if status == 'retrieved':
                    yield decode_result(key, result)
                    continue
                elif status == 'attached':
                    continue

            db.new_job(key, prov, job_msg)
            result = run_job(key, job)
            result_msg = registry.deep_encode(result.value)
            db.store_result(key, result_msg)
            yield result

    S = Scheduler()
    W = Queue() >> pass_job

    return S.run(W, get_workflow(wf))


def start_job(db):
    """Adds a time stamp to the database."""
    @pull
    def start_job_f(source):
        for key, job in source():
            db.add_time_stamp(key, 'start')
            yield key, job

    return start_job_f


@sink_map
def print_result(key, status, result, msg):
    print(status, result)


def schedule_job(results, registry, db,
                 job_keeper=None, pred=lambda job: True):
    """Schedule a job, providing there is no result for it in the database yet.

    First the database checks if there is a previous job that is identical to
    the current one. If this is the case, the result is 'retrieved'.

    If there is no result, but the job description is in the database, either
    the job is still running, or it was tried before but Noodles crashed.

    In the first case, the job is 'attached' to the already running job.
    In the second case, the record of the previous job is deleted and the new
    job is scheduled.
    """
    @push
    def schedule_f(job_sink_):
        job_sink = job_sink_()
        result_sink = results.sink()

        while True:
            key, job = yield

            if pred(job):
                job_msg = registry.deep_encode(job)
                prov = prov_key(job_msg)

                if db.job_exists(prov):
                    status, other_key, result = db.get_result_or_attach(
                        key, prov, job_keeper)
                    if status == 'retrieved':
                        result_sink.send(
                            (key, 'retrieved',
                             registry.deep_decode(result, deref=True), None))
                        continue
                    elif status == 'attached':
                        continue
                    elif status == 'broken':
                        db.new_job(key, prov, job_msg)
                        job_sink.send((key, job))

                else:
                    db.new_job(key, prov, job_msg)
                    job_sink.send((key, job))

            else:
                job_sink.send((key, job))

    return schedule_f


def store_result_deep(registry, db, job_keeper=None, pred=lambda job: True):
    """When the result is known, we can insert it in the database. This is
    only done if the result is not a workflow. If the result is a workflow,
    a 'link' is added in the database, identifying the workflow by the Python
    `id` of the workflow object. When this workflow is finished the final
    result is inserted in the database."""
    def store_result(key, result, msg):
        result_msg = registry.deep_encode(result)
        attached = db.store_result(key, result_msg)
        if attached:
            for akey in attached:
                yield ResultMessage(akey, 'attached', result, msg)

    @pull
    def f(source):
        for key, status, result, msg in source():
            wf, n = job_keeper[key]
            job = wf.nodes[n]

            if pred(job):
                if is_workflow(result):
                    db.add_link(key, id(get_workflow(result)))
                else:
                    yield from store_result(key, result, msg)

            if wf.root == n:
                linked_jobs = db.get_linked_jobs(id(wf))

                if is_workflow(result):
                    for k in linked_jobs:
                        db.add_link(k, id(get_workflow(result)))

                else:
                    for k in linked_jobs:
                        yield from store_result(k, result, msg)

            yield ResultMessage(key, status, result, msg)

    return f


def run_parallel(wf, n_threads, registry, jobdb_file, job_keeper=None):
    """Run a workflow in `n_threads` parallel threads. Now we replaced the
    single worker with a thread-pool of workers.

    This version works with the JobDB to cache results."""
    registry = registry()
    db = JobDB(jobdb_file)

    if job_keeper is None:
        job_keeper = JobKeeper()
    S = Scheduler(job_keeper=job_keeper)

    jobs = Queue()
    results = Queue()

    assert job_keeper is not None

    LogQ = Queue()
    threading.Thread(
        target=patch,
        args=(LogQ.source, job_keeper.message),
        daemon=True).start()

    @push_map
    def log_job_start(key, job):
        return (key, 'start', job, None)

    r_src = jobs.source \
        >> start_job(db) \
        >> branch(log_job_start >> LogQ.sink) \
        >> thread_pool(*repeat(worker, n_threads), results=results) \
        >> store_result_deep(registry, db, job_keeper) \
        >> branch(LogQ.sink)

    j_snk = schedule_job(results, registry, db, job_keeper) >> jobs.sink

    return S.run(Connection(r_src, j_snk), get_workflow(wf))


def create_prov_worker(
        worker, results, registry, jobdb_file, job_keeper,
        pred=lambda x: True, log_q=None):
    registry = registry()
    db = JobDB(jobdb_file)

    jobs = Queue()

    r_src = jobs.source \
        >> start_job(db) \
        >> worker \
        >> store_result_deep(registry, db, job_keeper, pred)

    @push_map
    def log_job_sched(key, job):
        return (key, 'schedule', job, None)

    j_snk = broadcast(
        log_job_sched >> log_q.sink,
        schedule_job(results, registry, db, job_keeper, pred) >> jobs.sink)

    return Connection(r_src, j_snk)


def prov_wrap_connection(
        worker, results, registry, jobdb_file, job_keeper,
        pred=lambda x: True, log_q=None):
    registry = registry()
    db = JobDB(jobdb_file)

    r_src = worker.source \
        >> store_result_deep(registry, db, job_keeper, pred)

    @push_map
    def log_job_sched(key, job):
        return (key, 'schedule', job, None)

    j_snk = broadcast(
        log_job_sched >> log_q.sink,
        schedule_job(results, registry, db, job_keeper, pred) >> worker.sink)

    return Connection(r_src, j_snk)


def run_parallel_opt(wf, n_threads, registry, jobdb_file,
                     job_keeper=None, display=None, cache_all=False):
    """Run a workflow in `n_threads` parallel threads. Now we replaced the
    single worker with a thread-pool of workers.

    This version works with the JobDB to cache results; however we only store
    the jobs that are hinted with the 'store' keyword.

    :param wf:
        The workflow.

    :param n_threads:
        The number of threads to start.

    :param registry:
        The serialisation to use in order to store results in the database,
        as well as identify jobs.

    :param jobdb_file:
        The filename of the job database.

    :param job_keeper:
        The JobKeeper instance to keep runtime information in. If not given,
        we create one for you and throw it in the trash when we're done.

    :param display:
        The display routine to display activity. If not given, we won't report
        on any activity.
    """
    if job_keeper is None:
        job_keeper = JobKeeper()
    S = Scheduler(job_keeper=job_keeper)

    results = Queue()

    if cache_all:
        def pred(job):
            return True
    else:
        def pred(job):
            return job.hints and 'store' in job.hints

    LogQ = Queue()
    if display:
        tgt = broadcast(job_keeper.message, sink_map(display))
    else:
        tgt = job_keeper.message

    threading.Thread(
        target=patch,
        args=(LogQ.source, tgt),
        daemon=True).start()

    parallel_worker = \
        thread_pool(*repeat(worker, n_threads), results=results) >> \
        branch(LogQ.sink)

    return S.run(
        create_prov_worker(
            parallel_worker, results, registry, jobdb_file, job_keeper,
            pred, LogQ),
        get_workflow(wf))



import pytest

from noodles import run_process, schedule, serial

try:
    import msgpack  # noqa
    has_msgpack = True
except ImportError:
    has_msgpack = False

@schedule
def writes_to_stdout():
    print("Hello Noodles!")
    return 42

@pytest.mark.skipif(not has_msgpack, reason="msgpack needed.")
def test_capture_output():
    a = writes_to_stdout()
    result = run_process(a, n_processes=1, registry=serial.base, use_msgpack=True)
    assert result == 42

def test_capture_output_nomsgpack():
    a = writes_to_stdout()
    result = run_process(a, n_processes=1, registry=serial.base, use_msgpack=False)
    assert result == 42



"""
The Noodles Runtime
===================

The Noodles runtime can be thought of as a series of components passing each
other messages. At the center of this system is the *scheduler*, which sends
out jobs to a worker and receives results.  This abstracted form of a worker is
implemented as a pair of coroutines.  The job-messages are a two-tuple, having
a key and a job object. The communication of results is done through
four-tuples, as `(key, status, result, message)`. The `key` must be a key
identifying with a previously submitted job. The `status` is one of
'scheduled', 'fizzled', 'started', 'done', 'aborted', 'error'. There can only
be a `result` if the status is `done`, otherwise `result` must be `None`. The
`message` could be an exception object, or any other object, or `None`. How to
treat the contents of a message is up to the *cook* (library implementation).

Handling of errors
~~~~~~~~~~~~~~~~~~

When the returned status is 'aborted', this means that there was some internal
error in the runtime; for example, a node at the cluster killed your job!  If a
thread is guarded against such things happening (an exception is raised) it
should notify the scheduler about all outstanding jobs on that node with an
'aborted' status. The logical thing to do for the scheduler is then to reschedule
these jobs on a different machine.

The 'fizzled' status means that the job could not start in the first place. Maybe
a resource that we need access to is unavailable. The default behaviour is not to
restart fizzled jobs unless a user intervenes.

The user has every option to catch errors himself, and change the program logic
accordingly. For instance, if we do a parameter sweep over a function that doesn't
converge or is undefined for some value, we may want to proceed with the rest of
the computation. It is the responsibility of the user (and the cook) to catch
all exceptions before they reach Noodles. If an exception does get through, the
worker will return a status 'error' and put the exception in the message field.
Then we can do three things: kill everything, wait for outstanding jobs to finish
(graceful exit), or
ignore that there was an error and put `None` for a result. In the last case the
exception should be a `Warning`. For all other situations the default behaviour is
the graceful exit. To stop all work being done the user can give a keyboard interupt.
"""




from .coroutine import coroutine

import inspect


def haploid(mode):
    def wrap(f):
        if mode == 'send':
            return Haploid(coroutine(f), mode)
        else:
            return Haploid(f, mode)

    return wrap


class Haploid(object):
    """A Haploid is a single co-routine sending or pulling data.

    A pulling haploid could have the form::

        @haploid('pull')
        def foo(source):
            for item in source:
                yield do_something(item)

    while a sending haploid  (recognizable from the `send` call) would look
    like::

        @haploid('send')
        def foo(sink):
            while True:
                item = yield
                sink.send(do_something(item))

    The difference between these two modes is all about ownership of execution.
    If we want to link a haploid with active output (a sender) to a puller
    having active input, it is assumed the data transfer is approached from two
    different thread points (either two different threads simultaneously, or
    the same thread at a different point of execution). The universal way of
    fixing this is by putting a Queue object in between.

    The other way around, having two passive ends, requires the insertion of a
    thread running the simple function:

        def patch(source, sink):
            for item in source:
                sink.send(item)
    """
    def __init__(self, fn, mode):
        self.fn = fn       # if mode == 'pull' else coroutine(fn)
        self.mode = mode   # 'send' or 'pull'
        if mode == 'send':
            self.mode = 'push'

    def __rshift__(self, other):
        return self.__join__(other)

    def __call__(self, *args):
        return self.fn(*args)


class pull(Haploid):
    def __init__(self, fn):
        self.source = fn
        super(pull, self).__init__(fn, 'pull')

    def __join__(self, other):
        if hasattr(other, 'source'):
            return pull(lambda *args: other.source(lambda: self.source(*args)))
        elif inspect.isfunction(other):
            return self >> pull_map(other)
        else:
            raise TypeError('Cannot join coroutines of different types.')


class push(Haploid):
    def __init__(self, fn, dont_wrap=False):
        if dont_wrap:
            self.sink = fn
        else:
            self.sink = coroutine(fn)
        super(push, self).__init__(self.sink, 'push')

    def __join__(self, other):
        if hasattr(other, 'sink'):
            return push(lambda *args: self.sink(lambda: other.sink(*args)),
                        dont_wrap=True)
        elif inspect.isfunction(other):
            return self >> push_map(other)
        else:
            raise TypeError('Cannot join coroutines of different types.')


def compose_2(f, g):
    def h(*args):
        a = g(*args)
        if a:
            return f(*a)
        else:
            return None

    return h


class composer:
    def __init__(self, f):
        self.f = f

    def __call__(self, *args):
        return self.f(*args)

    def __rshift__(self, other):
        return composer(compose_2(other, self.f))


class pull_map(pull):
    def __init__(self, f):
        self.f = f
        super(pull_map, self).__init__(self.map)

    def __join__(self, other):
        if not isinstance(other, Haploid) and inspect.isfunction(other):
            return pull_map(compose_2(other, self.f))

        if isinstance(other, pull_map):
            return pull_map(compose_2(other.f, self.f))

        else:
            return super(pull_map, self).__join__(other)

    def map(self, source):
        for args in source():
            if args:
                yield self.f(*args)
            else:
                yield


class push_map(push):
    def __init__(self, f):
        self.f = f
        super(push_map, self).__init__(self.map)

    def __join__(self, other):
        if not isinstance(other, Haploid) and inspect.isfunction(other):
            return push_map(compose_2(other, self.f))

        if isinstance(other, push_map):
            return push_map(compose_2(other.f, self.f))

        else:
            return super(push_map, self).__join__(other)

    def map(self, sink):
        sink = sink()
        while True:
            args = yield
            if args:
                sink.send(self.f(*args))
            else:
                raise RuntimeError('Encountered `None` in push_map loop.')
                # sink.send(None)


def sink_map(f):
    @push
    def sink():
        while True:
            args = yield
            if args:
                f(*args)
            else:
                raise RuntimeError('Encountered `None` in sink_map loop.')

    return sink


def branch(*sinks_):
    @pull
    def junction(source):
        sinks = [s() for s in sinks_]

        for msg in source():
            for s in sinks:
                s.send(msg)
            yield msg

    return junction


def broadcast(*sinks_):
    @push
    def bc():
        sinks = [s() for s in sinks_]
        while True:
            msg = yield
            for s in sinks:
                s.send(msg)

    return bc


def patch(source, sink):
    """Create a direct link between a source and a sink."""
    sink = sink()
    for v in source():
        if v is not None:
            sink.send(v)
        else:
            raise RuntimeError("encountered None in stream")



from pytest import raises
from noodles import (schedule, schedule_hint, run_single)


class MyException(Exception):
    def __init__(self, msg):
        super(MyException, self).__init__(msg)


@schedule
def raises_my_exception():
    raise MyException("Error!")


def test_exception_00():
    with raises(MyException):
        wf = raises_my_exception()
        run_single(wf)


@schedule_hint(annotated=True)
def not_really_annotated():
    return 0


def test_exception_01():
    with raises(TypeError):
        wf = not_really_annotated()
        run_single(wf)



import pytest

try:
    import numpy as np
    from numpy import (random, fft, exp)
    from noodles.serial.numpy import arrays_to_hdf5

    from noodles.run.run_with_prov import (
        run_parallel_opt)

except ImportError:
    has_numpy = False

else:
    has_numpy = True

from noodles.display import (NCDisplay)
from noodles import schedule, serial, schedule_hint


def registry():
    return serial.base() + arrays_to_hdf5()


@schedule_hint(display="fft", confirm=True, store=True)
def do_fft(a):
    return fft.fft(a)


@schedule
def make_kernel(n, sigma):
    return exp(-fft.fftfreq(n)**2 * sigma**2)


@schedule_hint(display="ifft", confirm=True, store=True)
def do_ifft(a):
    return fft.ifft(a).real


@schedule
def apply_filter(a, b):
    return a * b


@schedule_hint(display="make noise {seed}", confirm=True, store=True)
def make_noise(n, seed=0):
    random.seed(seed)
    return random.normal(0, 1, n)


def run(wf):
    with NCDisplay() as display:
        result = run_parallel_opt(
            wf, 2, registry, "cache.json", display=display)

    return result


@pytest.mark.skipif(not has_numpy, reason="NumPy needed.")
def test_hdf5():
    x = make_noise(256)
    k = make_kernel(256, 10)
    x_smooth = do_ifft(apply_filter(do_fft(x), k))
    result = run(x_smooth)

    assert isinstance(result, np.ndarray)
    assert result.size == 256



import sys
import uuid
from subprocess import Popen, PIPE
import threading

import os
import random
from ..workflow import get_workflow
# from ..logger import log
from .connection import Connection
from ..utility import object_name
from .scheduler import Scheduler
# from .protect import CatchExceptions
from .hybrid import hybrid_threaded_worker
from .haploid import (pull, push)

from .remote.io import (
    MsgPackObjectReader, MsgPackObjectWriter,
    JSONObjectReader, JSONObjectWriter)

try:
    import msgpack  # noqa
    has_msgpack = True
except ImportError:
    has_msgpack = False


def process_worker(registry, verbose=False, jobdirs=False,
                   init=None, finish=None, status=True, use_msgpack=False):
    name = "process-" + str(uuid.uuid4())

    cmd = ["/bin/bash", os.getcwd() + "/worker.sh", sys.prefix, "online",
           "-name", name, "-registry", object_name(registry)]
    if use_msgpack:
        assert has_msgpack
        cmd.append("-msgpack")
    if verbose:
        cmd.append("-verbose")
    if jobdirs:
        cmd.append("-jobdirs")
    if not status:
        cmd.append("-nostatus")
    if init:
        cmd.extend(["-init", object_name(init)])
    if finish:
        cmd.extend(["-finish", object_name(finish)])

    p = Popen(
        cmd,
        stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)

    def read_stderr():
        for line in p.stderr:
            print(name + ": " + line.strip(), file=sys.stderr, flush=True)

    t = threading.Thread(target=read_stderr)
    t.daemon = True
    t.start()

    @push
    def send_job():
        reg = registry()
        if use_msgpack:
            yield from MsgPackObjectWriter(reg, p.stdin.buffer)
        else:
            yield from JSONObjectWriter(reg, p.stdin)

    @pull
    def get_result():
        reg = registry()
        if use_msgpack:
            newin = os.fdopen(p.stdout.fileno(), 'rb', buffering=0)
            yield from MsgPackObjectReader(reg, newin)
        else:
            yield from JSONObjectReader(reg, p.stdout)

    return Connection(get_result, send_job, aux=read_stderr)


def run_process(wf, n_processes, registry,
                verbose=False, jobdirs=False,
                init=None, finish=None, deref=False, use_msgpack=False):
    """Run the workflow using a number of new python processes. Use this
    runner to test the workflow in a situation where data serial
    is needed.

    :param wf:
        The workflow.
    :type wf: `Workflow` or `PromisedObject`

    :param n_processes:
        Number of processes to start.

    :param registry:
        The serial registry.

    :param verbose:
        Request verbose output on worker side

    :param jobdirs:
        Create a new directory for each job to prevent filename collision.(NYI)

    :param init:
        An init function that needs to be run in each process before other jobs
        can be run. This should be a scheduled function returning True on
        success.

    :param finish:
        A function that wraps up when the worker closes down.

    :param deref:
        Set this to True to pass the result through one more encoding and
        decoding step with object derefencing turned on.
    :type deref: bool

    :returns: the result of evaluating the workflow
    :rtype: any
    """
    workers = {}
    for i in range(n_processes):
        new_worker = process_worker(registry, verbose, jobdirs, init, finish,
                                    use_msgpack=use_msgpack)
        workers['worker {0:2}'.format(i)] = new_worker

    worker_names = list(workers.keys())

    def random_selector(job):
        return random.choice(worker_names)

    master_worker = hybrid_threaded_worker(random_selector, workers)
    result = Scheduler().run(master_worker, get_workflow(wf))

    if deref:
        return registry().dereference(result, host='localhost')
    else:
        return result



from queue import Queue
from ..utility import (
    object_name, look_up, deep_map, inverse_deep_map)

import noodles

# try:
#    import ujson as json
# except ImportError:
import json

try:
    import msgpack
    has_msgpack = True
except ImportError:
    has_msgpack = False


def _chain_fn(a, b):
    def f(obj):
        first = a(obj)
        if first:
            return first

        return b(obj)

    return f


class RefObject:
    """Placeholder object to delay decoding a serialised object
    until needed by a worker."""
    def __init__(self, rec):
        self.rec = rec


class Registry(object):
    """Serialisation registry, keeps a record of `Serialiser` objects.

    The Registry keeps a dictionary mapping (qualified) class names to
    :py:class:`Serialiser` objects. Given an object, the `__getitem__`
    method looks for the highest base class that it has a serialiser for.
    As a fall-back we install a Serialiser matching the Python
    `object` class.

    Detection by object type is not always meaningful or even possible.
    Before scannning for known base classes the look-up function passes
    the object through the `hook` function, which should return a string
    or `None`. If a string is returned that string is used to look-up
    the serialiser.

    Registries can be combined using the '+' operator. The left side argument
    is than used as `parent` to the new Registry, while the right-hand argument
    overrides and augments the Serialisers present. The `hook` functions
    are being chained, such that the right-hand registry takes precedence.
    The default serialiser is inherrited from the left-hand argument.
    """
    def __init__(self, parent=None, types=None, hooks=None, hook_fn=None,
                 default=None):
        """Constructor

        :param parent:
            The new Registry takes the dictionary and hook from the parent.
            If no other argumentns are given, we get a copy of `parent`.
        :type parent: `Registry`

        :param types:
            A dictionary of types to Serialiser objects. Each of these
            are added to the new Registry.

        :param hooks:
            A dictionary of strings to Serialiser objects. These are added
            directly to the dictionary internal to the new Registry.

        :param hook_fn:
            A function taking an object returning a string. The string should
            match a string in the hook dictionary. It should not be possible
            to confuse the returned string with a qualified Python name.
            One way to do this, is by enclosing the string with
            '<' '>' characters.

        :param default:
            The default fall-back for the new Registry.
        :type default: `Serialiser`"""
        self._sers = parent._sers.copy() if parent else {}

        if types:
            for k, v in types.items():
                self[k] = v

        if hooks:
            self._sers.update(hooks)

        self.default = default if default \
            else parent.default if parent \
            else Serialiser(object)

        if hook_fn and parent and parent._hook:
            self._hook = _chain_fn(hook_fn, parent._hook)
        else:
            self._hook = hook_fn if hook_fn \
                else parent._hook if parent \
                else None

    def __add__(self, other):
        """Merge two registries. Right-side takes presedence over left-side
        argument, with exception of the default (fall-back) serialiser."""
        reg = Registry(
            parent=self, hooks=other._sers,
            hook_fn=other._hook, default=self.default)

        return reg

    @property
    def default(self):
        return self[object]

    @default.setter
    def default(self, ser):
        self[object] = ser

    def __getitem__(self, key):
        """Searches the most fitting serialiser based on the inheritance tree
        of the given class. We search this tree breadth-first."""
        q = Queue()  # use a queue for breadth-first decent

        q.put(key)
        while not q.empty():
            cls = q.get()
            m_n = object_name(cls)

            if m_n in self._sers:
                return self._sers[m_n]
            else:
                for base in cls.__bases__:
                    q.put(base)

    def __setitem__(self, cls, value):
        """Sets a new Serialiser for the given class."""
        m_n = object_name(cls)
        self._sers[m_n] = value

    def encode(self, obj, host=None):
        """Encode an object using the serialisers available
        in this registry. Objects that have a type that is one of
        [dict, list, str, int, float, bool, tuple] are send back unchanged.

        A host-name can be given as an additional argument to identify the
        host in the resulting record if the encoder yields any filenames.

        This function only treats the object for one layer deep.

        :param obj:
            The object that needs encoding.

        :param host:
            The name of the encoding host.
        :type host: str
        """
        if obj is None:
            return None

        if type(obj) in [dict, list, str, int, float, bool]:
            return obj

        if isinstance(obj, RefObject):
            return obj.rec

        hook = self._hook(obj) if self._hook else None
        typename = hook if hook else object_name(type(obj))

        def make_rec(data, ref=None, files=None):
            rec = {'_noodles': noodles.__version__,
                   'type': typename,
                   'data': data}

            if ref is not None:
                rec['ref'] = ref,

            if files:
                rec['host'] = host
                rec['files'] = files

            return rec

        if hook:
            return self._sers[hook].encode(obj, make_rec)

        enc = self[type(obj)]
        result = enc.encode(obj, make_rec)
        return result

    def decode(self, rec, deref=False):
        """Decode a record to return an object that could be considered
        equivalent to the original.

        The record is not touched if `_noodles` is not an item in the record.

        :param rec:
            A dictionary record to be decoded.
        :type rec: dict

        :param deref:
            Wether to decode a RefObject. If the encoder wrote files on a
            remote host, reading this file will be slow and result in an
            error if the file is not present.
        :type deref: bool"""
        if not isinstance(rec, dict):
            return rec

        if '_noodles' not in rec:
            return rec

        # if not deref:
        if rec.get('ref', False) and not deref:
            return RefObject(rec)

        typename = rec['type']
        if typename[0] == '<' and typename[-1] == '>':
            return self._sers[typename].decode(None, rec['data'])

        cls = look_up(typename)
        return self[cls].decode(cls, rec['data'])

    def deep_encode(self, obj, host=None):
        return deep_map(lambda o: self.encode(o, host), obj)

    def deep_decode(self, rec, deref=False):
        return inverse_deep_map(lambda r: self.decode(r, deref), rec)

    def to_json(self, obj, host=None, indent=None):
        """Recursively encode `obj` and convert it to a JSON string.

        :param obj:
            Object to encode.

        :param host:
            hostname where this object is being encoded.
        :type host: str"""
        if indent:
            return json.dumps(deep_map(lambda o: self.encode(o, host), obj),
                              indent=indent)
        else:
            return json.dumps(deep_map(lambda o: self.encode(o, host), obj))

    def to_msgpack(self, obj, host=None):
        return msgpack.packb(deep_map(lambda o: self.encode(o, host), obj))

    def from_json(self, data, deref=False):
        """Decode the string from JSON to return the original object (if
        `deref` is true. Uses the `json.loads` function with `self.decode`
        as object_hook.

        :param data:
            JSON encoded string.
        :type data: str

        :param deref:
            Whether to decode records that gave `ref=True` at encoding.
        :type deref: bool"""
        # return self.deep_decode(json.loads(data), deref)
        return json.loads(data, object_hook=lambda o: self.decode(o, deref))

    def from_msgpack(self, data, deref=False):
        return msgpack.unpackb(
            data,
            object_hook=lambda o: self.decode(o, deref))

    def dereference(self, data, host):
        """Dereferences RefObjects stuck in the hierarchy. This is a bit
        of an ugly hack."""
        return self.from_json(self.to_json(data, host), deref=True)


class Serialiser(object):
    """Serialiser base class.

    Serialisation classes should derive from `Serialiser` and overload the
    `encode` and `decode` methods.

    :param base:
        The type that this class is supposed to serialise. This may differ
        from the type of the object actually being serialised if its class
        was derived from `base`. The supposed base-class is kept here for
        reference but serves no immediate purpose.
    :type base: type"""
    def __init__(self, name):
        if isinstance(name, str):
            self.name = name
        else:
            try:
                self.name = name.__name__
            except AttributeError:
                self.name = '<unknown>'

    def encode(self, obj, make_rec):
        """Should encode an object of type `self.base` (or derived).

        This method receives the object and a function `make_rec`. This
        function has signature:

        .. code-block:: python

            def make_rec(rec, ref=False, files=None):
                ...

        If encoding and decoding is somewhat cosuming on resources, the
        encoder may call with `ref=True`. Then the resulting record won't
        be decoded until needed by the next job. This is most certainly
        the case when an external file was written. In this case the
        filename(s) should be passed as a list by `files=[...]`.

        The `files` list is not passed back to the decoder. Rather it is used
        by noodles to keep track of written files and copy them between hosts
        if needed. It is the responsibily of the encoder to include
        the filename information in the passed record as well.

        :param obj:
            Object to be encoded.

        :param make_rec:
            Function used to pack the encoded data with some meta-data."""
        if obj is None:
            raise RuntimeError("Object None should not reach encoder.")

        if hasattr(obj, '__serialize__'):
            return obj.__serialize__(make_rec)

        msg = "Cannot encode {}: encoder for type `{}` is not implemented." \
            .format(obj, type(obj).__name__)

        raise NotImplementedError(msg)

    @staticmethod
    def decode(cls, data):
        """Should decode the data to an object of type 'cls'.

        :param cls:
            The class is retrieved by the qualified name of the type
            of the object that was encoded; restored by importing it.
        :type cls: type

        :param data:
            The data is the record that was passed to `make_rec` by
            the encoder."""

        if hasattr(cls, '__construct__'):
            return cls.__construct__(data)

        msg = "Decoder for type `{}` is not implemented." \
            .format(cls.__name__)

        raise NotImplementedError(msg)
